# Copyright 2018 (c) Foundries.io.
#
# SPDX-License-Identifier: Apache-2.0

'''Common definitions for building Zephyr applications.

This provides some default settings and convenience wrappers for
building Zephyr applications needed by multiple commands.

See west.cmd.build for the build command itself.
'''

from west import cmake
from west import log

DEFAULT_BUILD_DIR = 'build'
'''Name of the default Zephyr build directory.'''

DEFAULT_CMAKE_GENERATOR = 'Ninja'
'''Name of the default CMake generator.'''


def is_zephyr_build(path):
    '''Return true if and only if `path` appears to be a valid Zephyr
    build directory.

    "Valid" means the given path is a directory which contains a CMake
    cache with a 'ZEPHYR_TOOLCHAIN_VARIANT' key.
    '''
    try:
        cache = cmake.CMakeCache.from_build_dir(path)
    except FileNotFoundError:
        cache = {}

    if 'ZEPHYR_TOOLCHAIN_VARIANT' in cache:
        log.dbg('{} is a zephyr build directory'.format(path),
                level=log.VERBOSE_EXTREME)
        return True
    else:
        log.dbg('{} is NOT a valid zephyr build directory'.format(path),
                level=log.VERBOSE_EXTREME)
        return False



# Copyright (c) 2018, Nordic Semiconductor ASA
# Copyright 2018, 2019 Foundries.io Ltd
#
# SPDX-License-Identifier: Apache-2.0

'''Parser and abstract data types for west manifests.

The main class is Manifest. The recommended method for creating a
Manifest instance is via its ``from_file()`` or ``from_data()`` helper
methods.

There are additionally Defaults, Remote, and Project types defined,
which represent the values by the same names in a west
manifest. (I.e. "Remote" represents one of the elements in the
"remote" sequence in the manifest, and so on.) Some Default values,
such as the default project revision, may be supplied by this module
if they are not present in the manifest data.'''

import collections
import errno
from functools import lru_cache
import os
import shutil
import shlex
import subprocess

import configparser
import pykwalify.core
import yaml
from pathlib import PurePath

from west import util, log
from west.backports import CompletedProcess
from west.configuration import config


#: Index in projects where the project with contains project manifest file is
#: located
MANIFEST_PROJECT_INDEX = 0

#: The name of the branch that points to the revision specified in the
#: manifest
MANIFEST_REV_BRANCH = 'manifest-rev'

#: A qualified reference to MANIFEST_REV_BRANCH, i.e.
#: refs/heads/manifest-rev.
QUAL_MANIFEST_REV_BRANCH = 'refs/heads/' + MANIFEST_REV_BRANCH


def manifest_path():
    '''Return the path to the manifest file.

    Raises: WestNotFound if called from outside of a west working directory,
    MalformedConfig if the configuration file is missing a manifest.path key,
    and FileNotFoundError if the manifest.path file doesn't exist.'''
    try:
        ret = os.path.join(util.west_topdir(),
                           config.get('manifest', 'path'),
                           'west.yml')
    except (configparser.NoOptionError, configparser.NoSectionError) as e:
        raise MalformedConfig('no "manifest.path" config option is set') from e
    # It's kind of annoying to manually instantiate a FileNotFoundError.
    # This seems to be the best way.
    if not os.path.isfile(ret):
        raise OSError(errno.ENOENT, os.strerror(errno.ENOENT), ret)
    return ret


class Manifest:
    '''Represents the contents of a West manifest file.

    The most convenient way to construct an instance is using the
    from_file and from_data helper methods.'''

    @staticmethod
    def from_file(source_file=None):
        '''Create and return a new Manifest object given a source YAML file.

        :param source_file: Path to a YAML file containing the manifest.

        If source_file is None, the value returned by `manifest_path()`
        is used.

        Raises `MalformedManifest` in case of validation errors.
        Raises `MalformedConfig` in case of missing configuration settings.'''
        if source_file is None:
            source_file = manifest_path()
        return Manifest(source_file=source_file)

    @staticmethod
    def from_data(source_data):
        '''Create and return a new Manifest object given parsed YAML data.

        :param source_data: Parsed YAML data as a Python object.

        Raises MalformedManifest in case of validation errors.
        Raises MalformedConfig in case of missing configuration settings.'''
        return Manifest(source_data=source_data)

    def __init__(self, source_file=None, source_data=None):
        '''Create a new Manifest object.

        :param source_file: Path to a YAML file containing the manifest.
        :param source_data: Parsed YAML data as a Python object.

        Normally, it is more convenient to use the `from_file` and
        `from_data` convenience factories than calling the constructor
        directly.

        Exactly one of the source_file and source_data parameters must
        be given.

        Raises MalformedManifest in case of validation errors.
        Raises MalformedConfig in case of missing configuration settings.'''
        if source_file and source_data:
            raise ValueError('both source_file and source_data were given')

        if source_file:
            with open(source_file, 'r') as f:
                self._data = yaml.safe_load(f.read())
            path = source_file
        else:
            self._data = source_data
            path = None

        self.path = path
        '''Path to the file containing the manifest, or None if created
        from data rather than the file system.'''

        if not self._data:
            self._malformed('manifest contains no data')

        if 'manifest' not in self._data:
            self._malformed('manifest contains no manifest element')
        data = self._data['manifest']

        try:
            pykwalify.core.Core(source_data=data,
                                schema_files=[_SCHEMA_PATH]).validate()
        except pykwalify.errors.SchemaError as se:
            self._malformed(se._msg, parent=se)

        self.defaults = None
        '''west.manifest.Defaults object representing default values
        in the manifest, either as specified by the user or west itself.'''

        self.remotes = None
        '''Sequence of west.manifest.Remote objects representing manifest
        remotes. Note that not all projects have a remote.'''

        self.projects = None
        '''Sequence of west.manifest.Project objects representing manifest
        projects.

        Each element's values are fully initialized; there is no need
        to consult the defaults field to supply missing values.

        Note: The index MANIFEST_PROJECT_INDEX in sequence will hold the
        project which contains the project manifest file.'''

        # Set up the public attributes documented above, as well as
        # any internal attributes needed to implement the public API.
        self._load(self._data)

    def get_remote(self, name):
        '''Get a manifest Remote, given its name.'''
        return self._remotes_dict[name]

    def get_projects(self, project_ids, allow_paths=True, only_cloned=False):
        '''Get a list of Projects in the manifest from given *project_ids*.

        If *project_ids* is empty, a list containing all the
        manifest's projects is returned. The manifest is present
        as a project in this list at *MANIFEST_PROJECT_INDEX*, and the
        other projects follow in the order they appear in the manifest
        file.

        Otherwise, projects in the returned list are in the same order
        as specified in *project_ids*.

        Either of these situations raises a ValueError:

        - One or more non-existent projects is in *project_ids*
        - only_cloned is True, and the returned list would have
          contained one or more uncloned projects

        On error, the *args* attribute of the ValueError is a 2-tuple,
        containing a list of unknown project_ids at index 0, and a
        list of uncloned Projects at index 1.

        :param project_ids: A sequence of projects, identified by name
                            (at first priority) or path (as a
                            fallback, when allow_paths=True).
        :param allow_paths: If False, project_ids must be a sequence of
                            project names only; paths are not allowed.
        :param only_cloned: If True, ValueError is raised if an
                            uncloned project would have been returned.

        '''
        projects = list(self.projects)
        unknown = []   # all project_ids which don't resolve to a Project
        uncloned = []  # if only_cloned, resolved Projects which aren't cloned
        ret = []       # result list of resolved Projects

        # If no project_ids are specified, use all projects.
        if not project_ids:
            if only_cloned:
                uncloned = [p for p in projects if not p.is_cloned()]
                if uncloned:
                    raise ValueError(unknown, uncloned)
            return projects

        # Otherwise, resolve each of the project_ids to a project,
        # returning the result or raising ValueError.
        for pid in project_ids:
            project = self._proj_name_map.get(pid)
            if project is None and allow_paths:
                project = self._proj_canon_path_map.get(util.canon_path(pid))

            if project is None:
                unknown.append(pid)
            else:
                ret.append(project)

                if only_cloned and not project.is_cloned():
                    uncloned.append(project)

        if unknown or (only_cloned and uncloned):
            raise ValueError(unknown, uncloned)
        return ret

    def as_frozen_dict(self):
        '''Returns an OrderedDict representing this manifest, frozen.

        The manifest is 'frozen' in that all Git revisions in the
        original data are replaced with the corresponding SHAs.

        Note that this requires that all projects are checked out.
        '''
        # Build a 'frozen' representation of all projects, except the
        # manifest project.
        projects = list(self.projects)
        del projects[MANIFEST_PROJECT_INDEX]
        frozen_projects = []
        for project in projects:
            if not project.is_cloned():
                raise RuntimeError('cannot freeze; project {} is uncloned'.
                                   format(project.name))
            sha = project.sha(QUAL_MANIFEST_REV_BRANCH)
            d = project.as_dict()
            d['revision'] = sha
            frozen_projects.append(d)

        # We include the defaults value here even though all projects
        # are fully specified in order to make the resulting manifest
        # easy to extend by users who want to reuse the defaults.
        r = collections.OrderedDict()
        r['manifest'] = collections.OrderedDict()
        r['manifest']['defaults'] = self.defaults.as_dict()
        r['manifest']['remotes'] = [r.as_dict() for r in self.remotes]
        r['manifest']['projects'] = frozen_projects
        r['manifest']['self'] = self.projects[MANIFEST_PROJECT_INDEX].as_dict()

        return r

    def _malformed(self, complaint, parent=None):
        context = (' file {} '.format(self.path) if self.path
                   else ' data:\n{}\n'.format(self._data))
        exc = MalformedManifest('Malformed manifest{}(schema: {}):\n{}'
                                .format(context, _SCHEMA_PATH,
                                        complaint))
        if parent:
            raise exc from parent
        else:
            raise exc

    def _load(self, data):
        # Initialize this instance's fields from values given in the
        # manifest data, which must be validated according to the schema.
        projects = []
        project_names = set()
        project_abspaths = set()

        manifest = data.get('manifest')

        path = config.get('manifest', 'path', fallback=None)

        self_tag = manifest.get('self')
        if path is None:
            path = self_tag.get('path') if self_tag else ''
        west_commands = self_tag.get('west-commands') if self_tag else None

        project = ManifestProject(path=path, west_commands=west_commands)
        projects.insert(MANIFEST_PROJECT_INDEX, project)

        # Map from each remote's name onto that remote's data in the manifest.
        remotes = tuple(Remote(r['name'], r['url-base']) for r in
                        manifest.get('remotes', []))
        remotes_dict = {r.name: r for r in remotes}

        # Get any defaults out of the manifest.
        #
        # md = manifest defaults (dictionary with values parsed from
        # the manifest)
        md = manifest.get('defaults', dict())
        mdrem = md.get('remote')
        if mdrem:
            # The default remote name, if provided, must refer to a
            # well-defined remote.
            if mdrem not in remotes_dict:
                self._malformed('default remote {} is not defined'.
                                format(mdrem))
            default_remote = remotes_dict[mdrem]
            default_remote_name = mdrem
        else:
            default_remote = None
            default_remote_name = None
        defaults = Defaults(remote=default_remote, revision=md.get('revision'))

        # mp = manifest project (dictionary with values parsed from
        # the manifest)
        for mp in manifest['projects']:
            # Validate the project name.
            name = mp['name']

            # Validate the project remote or URL.
            remote_name = mp.get('remote')
            url = mp.get('url')
            repo_path = mp.get('repo-path')
            if remote_name is None and url is None:
                if default_remote_name is None:
                    self._malformed(
                        'project {} has no remote or URL (no default is set)'.
                        format(name))
                else:
                    remote_name = default_remote_name
            if remote_name:
                if remote_name not in remotes_dict:
                    self._malformed('project {} remote {} is not defined'.
                                    format(name, remote_name))
                remote = remotes_dict[remote_name]
            else:
                remote = None

            # Create the project instance for final checking.
            try:
                project = Project(name,
                                  defaults,
                                  path=mp.get('path'),
                                  clone_depth=mp.get('clone-depth'),
                                  revision=mp.get('revision'),
                                  west_commands=mp.get('west-commands'),
                                  remote=remote,
                                  repo_path=repo_path,
                                  url=url)
            except ValueError as ve:
                self._malformed(ve.args[0])

            # Project names must be unique.
            if project.name in project_names:
                self._malformed('project name {} is already used'.
                                format(project.name))
            # Two projects cannot have the same path. We use absolute
            # paths to check for collisions to ensure paths are
            # normalized (e.g. for case-insensitive file systems or
            # in cases like on Windows where / or \ may serve as a
            # path component separator).
            if project.abspath in project_abspaths:
                self._malformed('project {} path {} is already in use'.
                                format(project.name, project.path))

            project_names.add(project.name)
            project_abspaths.add(project.abspath)
            projects.append(project)

        self.defaults = defaults
        self.remotes = remotes
        self._remotes_dict = remotes_dict
        self.projects = tuple(projects)
        self._proj_name_map = {p.name: p for p in self.projects}
        self._proj_canon_path_map = {util.canon_path(p.abspath): p
                                     for p in self.projects}


class MalformedManifest(Exception):
    '''Exception indicating that west manifest parsing failed due to a
    malformed value.'''


class MalformedConfig(Exception):
    '''Exception indicating that west config is malformed and thus causing west
       manifest parsing to fail.'''


# Definitions for Manifest attribute types.

class Defaults:
    '''Represents default values in a manifest, either specified by the
    user or by west itself.

    Defaults are neither comparable nor hashable.'''

    __slots__ = 'remote revision'.split()

    def __init__(self, remote=None, revision=None):
        '''Initialize a defaults value from manifest data.

        :param remote: Remote instance corresponding to the default remote,
                       or None (an actual Remote object, not the name of
                       a remote as a string).
        :param revision: Default Git revision; 'master' if not given.'''
        if remote is not None and not isinstance(remote, Remote):
            raise ValueError('{} is not a Remote'.format(remote))
        if revision is None:
            revision = 'master'

        self.remote = remote
        '''`Remote` corresponding to the default remote, or None.'''
        self.revision = revision
        '''Revision to applied to projects without an explicit value.'''

    def __eq__(self, other):
        return NotImplemented

    def __repr__(self):
        return 'Defaults(remote={}, revision={})'.format(repr(self.remote),
                                                         repr(self.revision))

    def as_dict(self):
        '''Return a representation of this object as a dict, as it would be
        parsed from an equivalent YAML manifest.'''
        ret = collections.OrderedDict()
        if self.remote and isinstance(self.remote, Remote):
            ret['remote'] = self.remote.name
        if self.revision:
            ret['revision'] = self.revision
        return ret


class Remote:
    '''Represents a remote defined in a west manifest.

    Remotes may be compared for equality, but are not hashable.'''

    __slots__ = 'name url_base'.split()

    def __init__(self, name, url_base):
        '''Initialize a remote from manifest data.

        :param name: remote's name
        :param url_base: remote's URL base.'''
        if url_base.endswith('/'):
            log.wrn('Remote', name, 'URL base', url_base,
                    'ends with a slash ("/"); these are automatically',
                    'appended by West')

        self.name = name
        '''Remote name as it appears in the manifest.'''
        self.url_base = url_base
        '''Remote url-base value as it appears in the manifest.'''

    def __eq__(self, other):
        return self.name == other.name and self.url_base == other.url_base

    def __repr__(self):
        return 'Remote(name={}, url-base={})'.format(repr(self.name),
                                                     repr(self.url_base))

    def as_dict(self):
        '''Return a representation of this object as a dict, as it would be
        parsed from an equivalent YAML manifest.'''
        return collections.OrderedDict(
            ((s.replace('_', '-'), getattr(self, s)) for s in self.__slots__))


class Project:
    '''Represents a project defined in a west manifest.

    Projects are neither comparable nor hashable.'''

    __slots__ = ('name remote url path abspath posixpath clone_depth '
                 'revision west_commands').split()

    def __init__(self, name, defaults=None, path=None, clone_depth=None,
                 revision=None, west_commands=None, remote=None,
                 repo_path=None, url=None):
        '''Specify a Project by name, Remote, and optional information.

        :param name: Project's user-defined name in the manifest.
        :param defaults: If the revision parameter is not given, the project's
                         revision is set to defaults.revision if defaults is
                         not None, or the west-wide default otherwise.
        :param path: Relative path to the project in the west
                     installation, if present in the manifest. If not given,
                     the project's ``name`` is used.
        :param clone_depth: Nonnegative integer clone depth if present in
                            the manifest.
        :param revision: Project revision as given in the manifest, if present.
                         If not given, defaults.revision is used instead.
        :param west_commands: path to a YAML file in the project containing
                              a description of west extension commands provided
                              by the project, if given.
        :param remote: Remote instance corresponding to this Project as
                       specified in the manifest. This is used to build
                       the project's URL, and is also stored as an attribute.
        :param repo_path: If this and *remote* are not None, then
                          ``remote.url_base + repo_path`` (instead of
                          ``remote.url_base + name``) is used as the project's
                          fetch URL.
        :param url: The project's fetch URL. This cannot be given with *remote*
                    or *repo_path*.
        '''
        if remote and url:
            raise ValueError('got both remote={} and url={}'.
                             format(remote, url))
        if repo_path and not remote:
            raise ValueError('got repo_path={} but no remote'.
                             format(repo_path, remote))
        if repo_path and url:
            raise ValueError('got both repo_path={} and url={}'.
                             format(repo_path, url))
        if not (remote or url):
            raise ValueError('got neither a remote nor a URL')

        if defaults is None:
            defaults = _DEFAULTS
        if repo_path is None:
            repo_path = name

        self.name = name
        '''Project name as it appears in the manifest.'''
        self.url = url or (remote.url_base + '/' + repo_path)
        '''Complete fetch URL for the project, either as given by the url kwarg
        or computed from the remote URL base and the project name.'''
        self.path = os.path.normpath(path or name)
        '''Relative path to the project in the installation.'''
        self.abspath = os.path.realpath(os.path.join(util.west_topdir(),
                                                     self.path))
        '''Absolute path to the project.'''
        self.posixpath = PurePath(self.abspath).as_posix()
        '''Absolute path to the project, POSIX style (with forward slashes).'''
        self.clone_depth = clone_depth
        '''Project's clone depth, or None'''
        self.revision = revision or defaults.revision
        '''Revision to check out for this project, as given in the manifest,
        from manifest defaults, or from the default supplied by west.'''
        self.west_commands = west_commands
        '''Path to project's "west-commands", or None.'''
        self.remote = remote
        '''`Remote` instance corresponding to the project's remote, or None.'''

    def __eq__(self, other):
        return NotImplemented

    def __repr__(self):
        reprs = [repr(x) for x in
                 (self.name, self.remote, self.url, self.path,
                  self.abspath, self.clone_depth, self.revision)]
        return ('Project(name={}, remote={}, url={}, path={}, abspath={}, '
                'clone_depth={}, revision={})').format(*reprs)

    def as_dict(self):
        '''Return a representation of this object as a dict, as it would be
        parsed from an equivalent YAML manifest.'''
        ret = collections.OrderedDict(
            (('name', self.name),
             ('remote', self.remote.name if self.remote else 'None'),
             ('revision', self.revision)))
        if self.path != self.name:
            ret['path'] = self.path
        if self.clone_depth:
            ret['clone-depth'] = self.clone_depth
        if self.west_commands:
            ret['west-commands'] = self.west_commands
        return ret

    def format(self, s, *args, **kwargs):
        '''Calls s.format() with instance-related format keys.

        The formatted value is returned.

        :param s: string (or other object) whose format() method to call

        The format method is called with ``*args`` and the following
        ``kwargs``:

        - this object's ``__slots__`` / values (name, url, etc.)
        - name_and_path: "self.name + (self.path)"
        - remote_name: "None" if no remote, otherwise self.remote.name
        - any additional kwargs passed as parameters

        The kwargs passed as parameters override the other values.
        '''
        kwargs = self._format_kwargs(kwargs)
        return s.format(*args, **kwargs)

    def _format_kwargs(self, kwargs):
        ret = {s: getattr(self, s) for s in self.__slots__}
        ret['name_and_path'] = '{} ({})'.format(self.name, self.path)
        ret['remote_name'] = ('None' if self.remote is None
                              else self.remote.name)
        ret.update(kwargs)
        return ret

    def git(self, cmd, extra_args=(), capture_stdout=False,
            capture_stderr=False, check=True, cwd=None):
        '''Helper for running a git command using metadata from a Project
        instance.

        :param cmd: git command as a string (or list of strings); all strings
                    are formatted using self.format() before use.
        :param extra_args: sequence of additional arguments to pass to the git
                           command (useful mostly if cmd is a string).
        :param capture_stdout: True if stdout should be captured into the
                               returned object instead of being printed.
        :param capture_stderr: Like capture_stdout, but for stderr. Use with
                               caution as it prevents error messages from being
                               shown to the user.
        :param check: True if a subprocess.CalledProcessError should be raised
                      if the git command finishes with a non-zero return code.
        :param cwd: directory to run command in (default: self.abspath)

        Returns a CompletedProcess (which is back-ported for Python 3.4).'''
        _warn_once_if_no_git()

        if isinstance(cmd, str):
            cmd_list = shlex.split(cmd)
        else:
            cmd_list = list(cmd)

        extra_args = list(extra_args)

        if cwd is None:
            cwd = self.abspath

        args = ['git'] + [self.format(arg) for arg in cmd_list] + extra_args
        cmd_str = util.quote_sh_list(args)

        log.dbg("running '{}'".format(cmd_str), 'in', cwd,
                level=log.VERBOSE_VERY)
        popen = subprocess.Popen(
            args, cwd=cwd,
            stdout=subprocess.PIPE if capture_stdout else None,
            stderr=subprocess.PIPE if capture_stderr else None)

        stdout, stderr = popen.communicate()

        dbg_msg = "'{}' in {} finished with exit status {}".format(
            cmd_str, cwd, popen.returncode)
        if capture_stdout:
            dbg_msg += " and wrote {} to stdout".format(stdout)
        if capture_stderr:
            dbg_msg += " and wrote {} to stderr".format(stderr)
        log.dbg(dbg_msg, level=log.VERBOSE_VERY)

        if check and popen.returncode:
            raise subprocess.CalledProcessError(popen.returncode, cmd_list,
                                                output=stdout, stderr=stderr)
        else:
            return CompletedProcess(popen.args, popen.returncode,
                                    stdout, stderr)

    def sha(self, rev):
        '''Returns the SHA of the given revision in the current project.

        :param rev: git revision (HEAD, v2.0.0, etc.) as a string
        '''
        cp = self.git('rev-parse ' + rev, capture_stdout=True)
        # Assumption: SHAs are hex values and thus safe to decode in ASCII.
        # It'll be fun when we find out that was wrong and how...
        return cp.stdout.decode('ascii').strip()

    def is_ancestor_of(self, rev1, rev2):
        '''Check if 'rev1' is an ancestor of 'rev2' in this project.

        :param rev1: commit that could be the ancestor
        :param rev2: commit that could be a descendant

        Returns True if rev1 is an ancestor commit of rev2 in the
        given project; rev1 and rev2 can be anything that resolves to
        a commit. (If rev1 and rev2 refer to the same commit, the
        return value is True, i.e. a commit is considered an ancestor
        of itself.) Returns False otherwise.'''
        returncode = self.git('merge-base --is-ancestor {} {}'.
                              format(rev1, rev2),
                              check=False).returncode

        if returncode == 0:
            return True
        elif returncode == 1:
            return False
        else:
            log.wrn(self.format(
                '{name_and_path}: git failed with exit code {rc}; '
                'treating as if "{r1}" is not an ancestor of "{r2}"',
                rc=returncode, r1=rev1, r2=rev2))
            return False

    def is_up_to_date_with(self, rev):
        '''Check if a project is up to date with revision 'rev'.

        :param rev: base revision to check if project is up to date with.

        Returns True if all commits in 'rev' are also in HEAD. This
        can be used to check if this project needs updates, rebasing,
        etc.; 'rev' can be anything that resolves to a commit.

        This is a special case of is_ancestor_of() provided for convenience.
        '''
        return self.is_ancestor_of(rev, 'HEAD')

    def is_up_to_date(self):
        '''Returns is_up_to_date_with(self.revision).'''
        return self.is_up_to_date_with(self.revision)

    def is_cloned(self):
        '''Returns True if the project's path is a directory that looks
        like the top-level directory of a Git repository, and False
        otherwise.'''
        if not os.path.isdir(self.abspath):
            return False

        # --is-inside-work-tree doesn't require that the directory is
        # the top-level directory of a Git repository. Use --show-cdup
        # instead, which prints an empty string (i.e., just a newline,
        # which we strip) for the top-level directory.
        res = self.git('rev-parse --show-cdup', check=False,
                       capture_stderr=True, capture_stdout=True)

        return not (res.returncode or res.stdout.strip())


class ManifestProject(Project):
    '''Represents the manifest as a project.'''

    def __init__(self, path=None, revision=None, url=None,
                 west_commands=None):
        '''Specify a Special Project by name, and url, and optional information.

        :param path: Relative path to the project in the west
                     installation, if present in the manifest. If None,
                     the project's ``name`` is used.
        :param revision: manifest project revision, or None
        :param url: Complete URL for the manifest project, or None
        :param west_commands: path to a YAML file in the project containing
                              a description of west extension commands provided
                              by the project, if given. This obviously only
                              makes sense for the manifest project, not west.
        '''
        self.name = path or 'manifest'
        '''Project's name (path or default "manifest").'''

        self.url = url
        '''Complete fetch URL for the project.'''

        self.path = path or self.name
        '''Relative path to the project in the installation.'''

        self.abspath = os.path.realpath(os.path.join(util.west_topdir(),
                                                     self.path))
        '''Absolute path to the project.'''

        self.posixpath = PurePath(self.abspath).as_posix()
        '''Absolute path to the project, POSIX style (with forward slashes).'''

        self.revision = revision
        '''Revision to check out for the west project, as given in the manifest,
        from manifest defaults, or from the default supplied by west. Undefined
        for the manifest project.'''

        self.remote = None
        '''None, provided for analogy with Project.'''

        self.clone_depth = None
        '''None, provided for analogy with Project.'''

        self.west_commands = west_commands
        '''Path to project's "west-commands", for the manifest project, or
        None.'''

    def as_dict(self):
        '''Return a representation of this object as a dict, as it would be
        parsed from an equivalent YAML manifest.'''
        ret = collections.OrderedDict({'path': self.path})
        if self.west_commands:
            ret['west-commands'] = self.west_commands
        return ret


_SCHEMA_PATH = os.path.join(os.path.dirname(__file__), "manifest-schema.yml")
_DEFAULTS = Defaults()


@lru_cache(maxsize=1)
def _warn_once_if_no_git():
    # Using an LRU cache means this gets called once. Afterwards, the
    # (nonexistent) memoized result is simply returned from the cache,
    # so the warning is emitted only once per process invocation.
    if shutil.which('git') is None:
        log.wrn('Git is not installed or cannot be found')



# Don't put anything else in here!
#
# This is the Python 3 version of option 3 in:
# https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version
__version__ = '0.6.2'



# Copyright 2018 Open Source Foundries Limited.
# Copyright 2019 Foundries.io Limited.
#
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from collections import OrderedDict
import importlib
import itertools
import os
import sys

import pykwalify
import yaml

from west import log
from west.configuration import config as _config
from west.manifest import Manifest
from west.util import escapes_directory

'''\
This package provides WestCommand, which is the common abstraction all
west commands subclass.

All built-in west commands are implemented as modules in this package.
This package also provides support for extension commands.'''

__all__ = ['CommandContextError', 'CommandError', 'WestCommand',
           'extension_commands', 'WestExtCommandSpec', 'ExtensionCommandError']

class CommandError(RuntimeError):
    '''Indicates that a command failed.'''

    def __init__(self, returncode=1):
        super().__init__()
        self.returncode = returncode

class CommandContextError(CommandError):
    '''Indicates that a context-dependent command could not be run.'''

class ExtensionCommandError(CommandError):
    '''Exception class indicating an extension command was badly
    defined and could not be created.'''

    def __init__(self, **kwargs):
        self.hint = kwargs.pop('hint', None)
        super(ExtensionCommandError, self).__init__(**kwargs)


_NO_TOPDIR_MSG_FMT = '''\
no west installation found from "{}"; "west {}" requires one.
Things to try:
 - Set ZEPHYR_BASE to a zephyr repository path in a west installation.
 - Run "west init" to set up an installation here.
 - Run "west init -h" for additional information.
'''

class WestCommand(ABC):
    '''Abstract superclass for a west command.'''

    def __init__(self, name, help, description, accepts_unknown_args=False,
                 requires_installation=True):
        '''Abstract superclass for a west command.

        Some of the fields in a WestCommand (such as *name*, *help*, and
        *description*) overlap with kwargs that should be passed to the
        argparse.ArgumentParser which should be added in the
        `WestCommand.add_parser()` method. This wart is by design: argparse
        doesn't make many API stability guarantees, so some of this
        information must be duplicated to work reliably in the future.

        :param name: the command's name, as entered by the user
        :param help: one-line command help text
        :param description: multi-line command description
        :param accepts_unknown_args: if true, the command can handle
                                     arbitrary unknown command line arguments
                                     in its run() method. Otherwise, passing
                                     unknown arguments will cause
                                     UnknownArgumentsError to be raised.
        :param requires_installation: if true, the command requires a
                                      west installation to run. Running such a
                                      command outside of any installation
                                      (i.e. without a valid topdir) will exit
                                      with an error.
        '''
        self.name = name
        self.help = help
        self.description = description
        self.accepts_unknown_args = accepts_unknown_args
        self.requires_installation = requires_installation
        self.topdir = None
        self.manifest = None

    def run(self, args, unknown, topdir, manifest=None):
        '''Run the command.

        This raises `CommandContextError` if the command cannot be run
        due to a context mismatch. Other exceptions may be raised as
        well.

        :param args: known arguments parsed via `WestCommand.add_parser()`.
        :param unknown: unknown arguments present on the command line;
                        this must be empty if the constructor
                        was passed accepts_unknown_args=False.
        :param topdir: top level directory of west installation, or None;
                       this is stored as an attribute before do_run() is
                       called.
        :param manifest: A pre-parsed west.manifest.Manifest, or None.
                         This will be saved in the self.manifest
                         field before do_run() is called.
        '''
        if unknown and not self.accepts_unknown_args:
            self.parser.error('unexpected arguments: {}'.format(unknown))
        if not topdir and self.requires_installation:
            log.die(_NO_TOPDIR_MSG_FMT.format(os.getcwd(), self.name))
        self.topdir = topdir
        self.manifest = manifest
        self.do_run(args, unknown)

    def add_parser(self, parser_adder):
        '''Registers a parser for this command, and returns it.

        The parser object is stored in the ``parser`` attribute of this
        WestCommand.

        :param parser_adder: The return value of a call to
                             argparse.ArgumentParser.add_subparsers()
        '''
        self.parser = self.do_add_parser(parser_adder)

        if self.parser is None:
            raise ValueError('no parser was returned')

        return self.parser

    #
    # Mandatory subclass hooks
    #

    @abstractmethod
    def do_add_parser(self, parser_adder):
        '''Subclass method for registering command line arguments.

        This is called by WestCommand.add_parser() to do the work of
        adding the parser itself.

        The subclass should call parser_adder.add_parser() to add an
        ArgumentParser for that subcommand, then add any
        arguments. The final parser must be returned.

        :param parser_adder: The return value of a call to
                             argparse.ArgumentParser.add_subparsers()

        '''

    @abstractmethod
    def do_run(self, args, unknown):
        '''Subclasses must implement; called when the command is run.

        :param args: is the namespace of parsed known arguments.
        :param unknown: If ``accepts_unknown_args`` was False when constructing
                        this object, this parameter is an empty sequence.
                        Otherwise, it is an iterable containing all unknown
                        arguments present on the command line.'''

    #
    # Public API, mostly for subclasses.
    #
    # These are meant to be useful to subclasses during their do_run()
    # calls. Using this functionality outside of a WestCommand
    # subclass leads to undefined results.
    #

    @property
    def manifest(self):
        '''Property for the manifest which was passed to run().

        If the manifest kwarg passed to do_run was not None, it is
        returned. Otherwise the access fails with a fatal error. This
        is meant to be used by a command to access the west manifest
        when it is required, and the command can't proceed
        successfully without it. The property is writeable.'''
        if self._manifest is None:
            log.die("can't run west {};".format(self.name),
                    "it requires the manifest, which was not available.",
                    'Try "west manifest --validate" to debug.')
        return self._manifest

    @manifest.setter
    def manifest(self, manifest):
        self._manifest = manifest

class WestExtCommandSpec:
    '''An object which allows instantiating an extension west command.'''

    def __init__(self, name, project, help, factory):
        self.name = name
        '''Command name, as known to the user.'''

        self.project = project
        '''west.manifest.Project instance which defined the command.'''

        self.help = help
        '''Help string in west-commands.yml, or a default value.'''

        self.factory = factory
        '''"Factory" callable for the command.

        This returns a WestCommand instance when called.
        It may do some additional steps (like importing the definition of
        the command) before constructing it, however.'''

def extension_commands(manifest=None):
    '''Get descriptions of available extension commands.

    The return value is an ordered map from project paths to lists of
    WestExtCommandSpec objects, for projects which define extension
    commands. The map's iteration order matches the manifest.projects
    order.

    The return value is empty if configuration option
    ``commands.allow_extensions`` is false.

    :param manifest: a parsed ``west.manifest.Manifest`` object, or None
                     to reload a new one.
    '''
    allow_extensions = _config.getboolean('commands', 'allow_extensions',
                                          fallback=True)
    if not allow_extensions:
        return {}

    if manifest is None:
        manifest = Manifest.from_file()

    specs = OrderedDict()
    for project in manifest.projects:
        if project.west_commands:
            specs[project.path] = _ext_specs(project)
    return specs

def _ext_specs(project):
    # Get a list of WestExtCommandSpec objects for the given
    # west.manifest.Project.

    spec_file = os.path.join(project.abspath, project.west_commands)

    # Verify project.west_commands isn't trying a directory traversal
    # outside of the project.
    if escapes_directory(spec_file, project.abspath):
        raise ExtensionCommandError(
            'west-commands file {} escapes project path {}'.
            format(project.west_commands, project.path))

    # Project may not be cloned yet.
    if not os.path.exists(spec_file):
        return []

    # Load the spec file and check the schema.
    with open(spec_file, 'r') as f:
        try:
            commands_spec = yaml.safe_load(f.read())
        except yaml.YAMLError as e:
            raise ExtensionCommandError from e
    try:
        pykwalify.core.Core(
            source_data=commands_spec,
            schema_files=[_EXT_SCHEMA_PATH]).validate()
    except pykwalify.errors.SchemaError as e:
        raise ExtensionCommandError from e

    ret = []
    for commands_desc in commands_spec['west-commands']:
        ret.extend(_ext_specs_from_desc(project, commands_desc))
    return ret

def _ext_specs_from_desc(project, commands_desc):
    py_file = os.path.join(project.abspath, commands_desc['file'])

    # Verify the YAML's python file doesn't escape the project directory.
    if escapes_directory(py_file, project.abspath):
        raise ExtensionCommandError(
            'extension command python file "{}" escapes project path {}'.
            format(commands_desc['file'], project.path))

    # Create the command thunks.
    thunks = []
    for command_desc in commands_desc['commands']:
        name = command_desc['name']
        attr = command_desc.get('class', name)
        help = command_desc.get(
            'help',
            '(no help provided; try "west {} -h")'.
            format(name))
        factory = _ExtFactory(py_file, name, attr)
        thunks.append(WestExtCommandSpec(name, project, help, factory))

    # Return the thunks for this project.
    return thunks

def _commands_module_from_file(file):
    # Python magic for importing a module containing west extension
    # commands. To avoid polluting the sys.modules key space, we put
    # these modules in an (otherwise unpopulated) west.commands.ext
    # package.
    #
    # The file is imported as a module named
    # west.commands.ext.A_FRESH_IDENTIFIER. This module object is
    # returned from a cache if the same file is ever imported again,
    # to avoid a double import in case the file maintains module-level
    # state or defines multiple commands.
    global _EXT_MODULES_CACHE
    global _EXT_MODULES_NAME_IT

    file = os.path.normcase(os.path.realpath(file))
    if file in _EXT_MODULES_CACHE:
        return _EXT_MODULES_CACHE[file]

    mod_name = next(_EXT_MODULES_NAME_IT)
    # The Python 3.4 way to import a module given its file got deprecated
    # later on, but we still need to support 3.4. If that requirement ever
    # gets dropped, this code can be simplified.
    if (3, 4) <= sys.version_info < (3, 5):
        def _import_mod_from(mod_name, file):
            from importlib.machinery import SourceFileLoader
            return SourceFileLoader(mod_name, file).load_module()
    else:
        def _import_mod_from(mod_name, file):
            spec = importlib.util.spec_from_file_location(mod_name, file)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
            return mod

    mod = _import_mod_from(mod_name, file)
    _EXT_MODULES_CACHE[file] = mod

    return mod


_EXT_SCHEMA_PATH = os.path.join(os.path.dirname(__file__),
                                'west-commands-schema.yml')

# Cache which maps files implementing extension commands their
# imported modules.
_EXT_MODULES_CACHE = {}
# Infinite iterator of "fresh" extension command module names.
_EXT_MODULES_NAME_IT = ('west.commands.ext.cmd_{}'.format(i)
                        for i in itertools.count(1))

class _ExtFactory:

    def __init__(self, py_file, name, attr):
        self.py_file = py_file
        self.name = name
        self.attr = attr

    def __call__(self):
        # Append the python file's directory to sys.path. This lets
        # its code import helper modules in a natural way.
        py_dir = os.path.dirname(self.py_file)
        sys.path.append(py_dir)

        # Load the module containing the command. Convert only
        # expected exceptions to ExtensionCommandError.
        try:
            mod = _commands_module_from_file(self.py_file)
        except ImportError as ie:
            raise ExtensionCommandError(
                hint='could not import {}'.format(self.py_file)) from ie

        # Get the attribute which provides the WestCommand subclass.
        try:
            cls = getattr(mod, self.attr)
        except AttributeError as ae:
            hint = 'no attribute {} in {}'.format(self.attr,
                                                  self.py_file)
            raise ExtensionCommandError(hint=hint) from ae

        # Create the command instance and return it.
        try:
            return cls()
        except Exception as e:
            raise ExtensionCommandError(
                hint='command constructor threw an exception') from e



#!/usr/bin/env python3

# Copyright 2018 Open Source Foundries Limited.
# Copyright 2019 Foundries.io Limited.
# Copyright (c) 2019, Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0

'''Zephyr RTOS meta-tool (west) main module
'''


import argparse
import colorama
from functools import partial
from io import StringIO
import itertools
import logging
import os
import shutil
import sys
from subprocess import CalledProcessError
import tempfile
import textwrap
import traceback

from west import log
from west import configuration as config
from west.commands import extension_commands, \
    CommandError, CommandContextError, ExtensionCommandError
from west.commands.project import List, ManifestCommand, Diff, Status, \
    SelfUpdate, ForAll, Init, Update
from west.commands.config import Config
from west.manifest import Manifest, MalformedConfig, MalformedManifest
from west.util import quote_sh_list, west_topdir, WestNotFound
from west.version import __version__

BUILTIN_COMMANDS = {
    'commands for managing multiple git repositories': [
        Init(),
        Update(),
        List(),
        ManifestCommand(),
        Diff(),
        Status(),
        ForAll(),
    ],

    'configuring west': [Config()],

    # None is for hidden commands we don't want to show to the user.
    None: [SelfUpdate()]
}

# Initialize the set with the virtual 'help' command so that an extension
# command cannot clash with it
BUILTIN_COMMAND_NAMES = set(['help'])
for group, commands in BUILTIN_COMMANDS.items():
    BUILTIN_COMMAND_NAMES.update(c.name for c in commands)


class WestHelpAction(argparse.Action):

    def __init__(self, option_strings, dest, **kwargs):
        kwargs['nargs'] = 0
        super(WestHelpAction, self).__init__(option_strings, dest,
                                             **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help(top_level=True)
        parser.exit()


class WestArgumentParser(argparse.ArgumentParser):
    # The argparse module is infuriatingly coy about its parser and
    # help formatting APIs, marking almost everything you need to
    # customize help output an "implementation detail". Even accessing
    # the parser's description and epilog attributes as we do here is
    # technically breaking the rules.
    #
    # Even though the implementation details have been pretty stable
    # since the module was first introduced in Python 3.2, let's avoid
    # possible headaches by overriding some "proper" argparse APIs
    # here instead of monkey-patching the module or breaking
    # abstraction barriers. This is duplicative but more future-proof.

    def __init__(self, *args, **kwargs):
        # The super constructor calls add_argument(), so this has to
        # come first as our override of that method relies on it.
        self.west_optionals = []
        self.west_extensions = None

        super(WestArgumentParser, self).__init__(*args, **kwargs)

    def print_help(self, file=None, top_level=False):
        print(self.format_help(top_level=top_level),
              file=file or sys.stdout)

    def format_help(self, top_level=False):
        # When top_level is True, we override the parent method to
        # produce more readable output, which separates commands into
        # logical groups. In order to print optionals, we rely on the
        # data available in our add_argument() override below.
        #
        # If top_level is False, it's because we're being called from
        # one of the subcommand parsers, and we delegate to super.

        if not top_level:
            return super(WestArgumentParser, self).format_help()

        # Format the help to be at most 75 columns wide, the maximum
        # generally recommended by typographers for readability.
        #
        # If the terminal width (COLUMNS) is less than 75, use width
        # (COLUMNS - 2) instead, unless that is less than 30 columns
        # wide, which we treat as a hard minimum.
        width = min(75, max(shutil.get_terminal_size().columns - 2, 30))

        with StringIO() as sio:

            def append(*strings):
                for s in strings:
                    print(s, file=sio)

            append(self.format_usage(),
                   self.description,
                   '')

            append('optional arguments:')
            for wo in self.west_optionals:
                self.format_west_optional(append, wo, width)

            append('')
            for group, commands in BUILTIN_COMMANDS.items():
                if group is None:
                    # Skip hidden commands.
                    continue

                append(group + ':')
                for command in commands:
                    self.format_command(append, command, width)
                append('')

            if self.west_extensions is None:
                # This only happens when there is an error.
                # If there are simply no extensions, it's an empty dict.
                append('Cannot load extension commands; '
                       'help for them is not available.')
                append('(To debug, try: "west manifest --validate".)')
                append('')
            else:
                # TODO we may want to be more aggressive about loading
                # command modules by default: the current implementation
                # prevents us from formatting one-line help here.
                #
                # Perhaps a commands.extension_paranoid that if set, uses
                # thunks, and otherwise just loads the modules and
                # provides help for each command.
                #
                # This has its own wrinkle: we can't let a failed
                # import break the built-in commands.
                for path, specs in self.west_extensions.items():
                    # This may occur in case a project defines commands already
                    # defined, in which case it has been filtered out.
                    if not specs:
                        continue

                    append('commands from project at "{}":'.
                           format(path))

                    for spec in specs:
                        self.format_extension_spec(append, spec, width)
                    append('')

            append(self.epilog)

            return sio.getvalue().rstrip()

    def format_west_optional(self, append, wo, width):
        metavar = wo['metavar']
        options = wo['options']
        help = wo.get('help')

        # Join the various options together as a comma-separated list,
        # with the metavar if there is one. That's our "thing".
        if metavar is not None:
            opt_str = '  ' + ', '.join('{} {}'.format(o, metavar)
                                       for o in options)
        else:
            opt_str = '  ' + ', '.join(options)

        # Delegate to the generic formatter.
        self.format_thing_and_help(append, opt_str, help, width)

    def format_command(self, append, command, width):
        thing = '  {}:'.format(command.name)
        self.format_thing_and_help(append, thing, command.help, width)

    def format_extension_spec(self, append, spec, width):
        self.format_thing_and_help(append, '  ' + spec.name + ':',
                                   spec.help, width)

    def format_thing_and_help(self, append, thing, help, width):
        # Format help for some "thing" (arbitrary text) and its
        # corresponding help text an argparse-like way.
        help_offset = min(max(10, width - 20), 24)
        help_indent = ' ' * help_offset

        thinglen = len(thing)

        if help is None:
            # If there's no help string, just print the thing.
            append(thing)
        else:
            # Reflow the lines in help to the desired with, using
            # the help_offset as an initial indent.
            help = ' '.join(help.split())
            help_lines = textwrap.wrap(help, width=width,
                                       initial_indent=help_indent,
                                       subsequent_indent=help_indent)

            if thinglen > help_offset - 1:
                # If the "thing" (plus room for a space) is longer
                # than the initial help offset, print it on its own
                # line, followed by the help on subsequent lines.
                append(thing)
                append(*help_lines)
            else:
                # The "thing" is short enough that we can start
                # printing help on the same line without overflowing
                # the help offset, so combine the "thing" with the
                # first line of help.
                help_lines[0] = thing + help_lines[0][thinglen:]
                append(*help_lines)

    def add_argument(self, *args, **kwargs):
        # Track information we want for formatting help.  The argparse
        # module calls kwargs.pop(), so can't call super first without
        # losing data.
        optional = {'options': [], 'metavar': kwargs.get('metavar', None)}
        need_metavar = (optional['metavar'] is None and
                        kwargs.get('action') in (None, 'store'))
        for arg in args:
            if not arg.startswith('-'):
                break
            optional['options'].append(arg)
            # If no metavar was given, the last option name is
            # used. By convention, long options go last, so this
            # matches the default argparse behavior.
            if need_metavar:
                optional['metavar'] = arg.lstrip('-').translate(
                    {ord('-'): '_'}).upper()
        optional['help'] = kwargs.get('help')
        self.west_optionals.append(optional)

        # Let argparse handle the actual argument.
        super(WestArgumentParser, self).add_argument(*args, **kwargs)

    def set_extensions(self, extensions):
        self.west_extensions = extensions


def _make_parsers():
    # Make a fresh instance of the top level argument parser,
    # subparser generator, and return them in that order.

    # The prog='west' override avoids the absolute path of the main.py script
    # showing up when West is run via the wrapper
    parser = WestArgumentParser(
        prog='west', description='The Zephyr RTOS meta-tool.',
        epilog='''Run "west <command> -h" or "west help <command>" for detailed
help on each command.''',
        add_help=False)

    # Remember to update scripts/west-completion.bash if you add or remove
    # flags

    parser.add_argument('-h', '--help', action=WestHelpAction,
                        help='show this help message and exit')

    parser.add_argument('-z', '--zephyr-base', default=None,
                        help='''Override the Zephyr base directory. The
                        default is the manifest project with path
                        "zephyr".''')

    parser.add_argument('-v', '--verbose', default=0, action='count',
                        help='''Display verbose output. May be given
                        multiple times to increase verbosity.''')

    parser.add_argument('-V', '--version', action='version',
                        version='West version: v{}'.format(__version__),
                        help='print the program version and exit')

    subparser_gen = parser.add_subparsers(metavar='<command>', dest='command')

    return parser, subparser_gen


def command_handler(command, topdir, manifest, known_args, unknown_args):
    command.run(known_args, unknown_args, topdir, manifest=manifest)


def add_ext_command_parser(subparser_gen, spec):
    # This subparser exists only to register the name. The real parser
    # will be created as needed later if the command is invoked. We
    # prevent help from being added because the default help printer
    # calls sys.exit(), which is not what we want.
    parser = subparser_gen.add_parser(spec.name, add_help=False)
    return parser


def ext_command_handler(spec, topdir, argv, manifest, *ignored):
    # Deferred creation, argument parsing, and handling for extension
    # commands. We go to the extra effort because we don't want to
    # import any extern classes until the user has specifically
    # requested an extension command.
    #
    # 'ignored' is just the known and unknown args as parsed by the
    # 'dummy' parser added by add_ext_command_parser().
    #
    # The purpose of this handler is to create the "real" parser that
    # we need for the newly instantiated `command`, then re-parse the
    # original argv and run the command.
    command = spec.factory()

    # Our original top level parser and subparser generator have some
    # garbage state that prevents us from registering the 'real'
    # command subparser. Just make new ones.
    west_parser, subparser_gen = _make_parsers()
    command.add_parser(subparser_gen)

    # Handle the instantiated command in the usual way.
    args, unknown = west_parser.parse_known_args(argv)
    command.run(args, unknown, topdir, manifest=manifest)


def help_command_handler(west_parser, topdir, help_parser, extensions, args,
                         *ignored):
    command_name = args.command
    if not command_name:
        west_parser.print_help(top_level=True)
        return

    if command_name in BUILTIN_COMMAND_NAMES:
        if command_name == 'help':
            help_parser.print_help()
            return
        for group, commands in BUILTIN_COMMANDS.items():
            for command in commands:
                if command.name == command_name:
                    command.parser.print_help()
                    return
    else:
        for path, specs in extensions.items():
            for spec in specs:
                if spec.name != command_name:
                    continue
                # ext_command_handler() does not return
                ext_command_handler(spec, topdir, [command_name, '--help'])
        else:
            west_parser.print_help(top_level=True)


def set_zephyr_base(args):
    '''Ensure ZEPHYR_BASE is set
    Order of precedence:
    1) Value given as command line argument
    2) Value from environment setting: ZEPHYR_BASE
    3) Value of zephyr.base setting in west config file

    Order of precedence between 2) and 3) can be changed with the setting
    zephyr.base-prefer.
    zephyr.base-prefer takes the values 'env' and 'configfile'

    If 2) and 3) has different values and zephyr.base-prefer is unset a warning
    is printed to the user.'''

    if args.zephyr_base:
        # The command line --zephyr-base takes precedence over
        # everything else.
        zb = os.path.abspath(args.zephyr_base)
        zb_origin = 'command line'
    else:
        # If the user doesn't specify it concretely, then use ZEPHYR_BASE
        # from the environment or zephyr.base from west.configuration.
        #
        # `west init` will configure zephyr.base to the project that has path
        # 'zephyr'.
        #
        # At some point, we need a more flexible way to set environment
        # variables based on manifest contents, but this is good enough
        # to get started with and to ask for wider testing.
        zb_env = os.environ.get('ZEPHYR_BASE')
        zb_prefer = config.config.get('zephyr', 'base-prefer',
                                      fallback=None)
        zb_config = config.config.get('zephyr', 'base', fallback=None)
        if zb_config is not None:
            zb_config = os.path.join(west_topdir(), zb_config)

        if zb_prefer == 'env' and zb_env is not None:
            zb = zb_env
            zb_origin = 'env'
        elif zb_prefer == 'configfile' and zb_config is not None:
            zb = zb_config
            zb_origin = 'configfile'
        elif zb_env is not None:
            zb = zb_env
            zb_origin = 'env'
            try:
                different = (zb_config is not None and
                             not os.path.samefile(zb_config, zb_env))
            except FileNotFoundError:
                different = (zb_config is not None and
                             (os.path.normpath(os.path.abspath(zb_config)) !=
                              os.path.normpath(os.path.abspath(zb_env))))
            if different:
                # The environment ZEPHYR_BASE takes precedence over the config
                # setting, but in normal multi-repo operation we shouldn't
                # expect to need to set ZEPHYR_BASE.
                # Therefore issue a warning as it might have happened that
                # zephyr-env.sh/cmd was run in some other zephyr installation,
                # and the user forgot about that.
                log.wrn('ZEPHYR_BASE={}'.format(zb_env),
                        'in the calling environment will be used, but was set '
                        'to', zb_config, 'in west config.\n To disable this '
                        'warning, execute '
                        '\'west config --global zephyr.base-prefer env\'')
        elif zb_config is not None:
            zb = zb_config
            zb_origin = 'configfile'
        else:
            zb = None
            zb_origin = None
            # No --zephyr-base, no ZEPHYR_BASE envronment and no zephyr.base
            # Fallback to loop over projects, to identify if a project has path
            # 'zephyr' for fallback.
            try:
                manifest = Manifest.from_file()
                for project in manifest.projects:
                    if project.path == 'zephyr':
                        zb = project.abspath
                        zb_origin = 'manifest file {}'.format(manifest.path)
                        break
                else:
                    log.err('no --zephyr-base given, ZEPHYR_BASE is unset,',
                            'west config contains no zephyr.base setting,',
                            'and no manifest project has path "zephyr"')
            except MalformedConfig as e:
                log.wrn("Can't set ZEPHYR_BASE:",
                        'parsing of manifest file failed during command',
                        args.command, ':', *e.args)
            except WestNotFound:
                log.wrn("Can't set ZEPHYR_BASE:",
                        'not currently in a west installation')

    if zb is not None:
        os.environ['ZEPHYR_BASE'] = zb
        log.dbg('ZEPHYR_BASE={} (origin: {})'.format(zb, zb_origin))


def parse_args(argv, extensions, topdir, manifest):
    west_parser, subparser_gen = _make_parsers()

    # Add handlers for the built-in commands.
    for command in itertools.chain(*BUILTIN_COMMANDS.values()):
        parser = command.add_parser(subparser_gen)
        parser.set_defaults(handler=partial(command_handler, command, topdir,
                                            manifest))

    # Add handlers for extension commands, and finalize the list with
    # our parser.
    if extensions:
        for path, specs in extensions.items():
            for spec in specs:
                parser = add_ext_command_parser(subparser_gen, spec)
                parser.set_defaults(handler=partial(ext_command_handler,
                                                    spec, topdir, argv,
                                                    manifest))
    west_parser.set_extensions(extensions)

    help_parser = subparser_gen.add_parser('help',
                                           help='get help on a west command')
    help_parser.add_argument('command', nargs='?')
    help_parser.set_defaults(handler=partial(help_command_handler, west_parser,
                                             topdir, help_parser, extensions))

    # Parse arguments.
    args, unknown = west_parser.parse_known_args(args=argv)

    # Set up logging verbosity before running the command, so
    # e.g. verbose messages related to argument handling errors work
    # properly. This works even for extension commands that haven't
    # been instantiated yet, because --verbose is an option to the top
    # level parser, and the command run() method doesn't get called
    # until later.
    log.set_verbosity(args.verbose)

    if args.command != 'init':  # init --> we haven't got a zephyr repo.
        set_zephyr_base(args)

    if 'handler' not in args:
        west_parser.print_help(file=sys.stderr, top_level=True)
        sys.exit(1)

    return args, unknown


def get_extension_commands(manifest):
    extensions = extension_commands(manifest=manifest)
    extension_names = set()

    for path, specs in extensions.items():
        # Filter out attempts to shadow built-in commands as well as
        # commands which have names which are already used.
        filtered = []
        for spec in specs:
            if spec.name in BUILTIN_COMMAND_NAMES:
                log.wrn('ignoring project {} extension command {};'.
                        format(spec.project.name, spec.name),
                        'this is a built in command')
                continue
            if spec.name in extension_names:
                log.wrn('ignoring project {} extension command "{}";'.
                        format(spec.project.name, spec.name),
                        'command "{}" already defined as extension command'.
                        format(spec.name))
                continue
            filtered.append(spec)
            extension_names.add(spec.name)
        extensions[path] = filtered

    return extensions


def dump_traceback():
    # Save the current exception to a file and return its path.
    fd, name = tempfile.mkstemp(prefix='west-exc-', suffix='.txt')
    os.close(fd)        # traceback has no use for the fd
    with open(name, 'w') as f:
        traceback.print_exc(file=f)
    return name


def main(argv=None):
    # Silence validation errors from pykwalify, which are logged at
    # logging.ERROR level. We want to handle those ourselves as
    # needed.
    logging.getLogger('pykwalify').setLevel(logging.CRITICAL)

    # Makes ANSI color escapes work on Windows, and strips them when
    # stdout/stderr isn't a terminal
    colorama.init()

    # See if we're in an installation.
    try:
        topdir = west_topdir()
    except WestNotFound:
        topdir = None

    # Read the configuration files before looking for extensions.
    # We need this to find the manifest path in order to load extensions.
    config.read_config()

    # Parse the manifest and create extension command thunks. We'll
    # pass the saved manifest around so it doesn't have to be
    # re-parsed.
    if topdir:
        try:
            manifest = Manifest.from_file()
            extensions = get_extension_commands(manifest)
        except (MalformedManifest, MalformedConfig, FileNotFoundError):
            manifest = None
            extensions = None
    else:
        manifest = None
        extensions = {}

    if argv is None:
        argv = sys.argv[1:]
    args, unknown = parse_args(argv, extensions, topdir, manifest)

    try:
        args.handler(args, unknown)
    except KeyboardInterrupt:
        sys.exit(0)
    except CalledProcessError as cpe:
        log.err('command exited with status {}: {}'.
                format(cpe.returncode, quote_sh_list(cpe.cmd)))
        if args.verbose:
            traceback.print_exc()
        sys.exit(cpe.returncode)
    except ExtensionCommandError as ece:
        msg = 'extension command "{}" could not be run{}.'.format(
            args.command, ': ' + ece.hint if ece.hint else '')
        if args.verbose:
            log.err(msg)
            traceback.print_exc()
        else:
            log.err(msg, 'See {} for a traceback.'.format(dump_traceback()))
        sys.exit(ece.returncode)
    except CommandContextError as cce:
        log.err('command', args.command, 'cannot be run in this context:',
                *cce.args)
        log.err('see {} for a traceback.'.format(dump_traceback()))
        sys.exit(cce.returncode)
    except CommandError as ce:
        # No need to dump_traceback() here. The command is responsible
        # for logging its own errors.
        sys.exit(ce.returncode)
    except (MalformedManifest, MalformedConfig) as malformed:
        log.die("can't load west manifest:", malformed)


if __name__ == "__main__":
    main()



# Copyright 2018 Open Source Foundries Limited.
# Copyright 2019 Foundries.io Limited.
#
# SPDX-License-Identifier: Apache-2.0

'''Provides common methods for printing messages to display to the user.

WestCommand instances should generally use the functions in this
module rather than calling print() directly if possible, as these
respect the ``color.ui`` configuration option and verbosity level.
'''

from west import configuration as config

import colorama
import sys

VERBOSE_NONE = 0
'''Default verbosity level, no dbg() messages printed.'''

VERBOSE_NORMAL = 1
'''Some verbose messages printed.'''

VERBOSE_VERY = 2
'''Very verbose output messages will be printed.'''

VERBOSE_EXTREME = 3
'''Extremely verbose output messages will be printed.'''

VERBOSE = VERBOSE_NONE
'''Global verbosity level. VERBOSE_NONE is the default.'''


def set_verbosity(value):
    '''Set the logging verbosity level.

    :param value: verbosity level to set, e.g. VERBOSE_VERY.
    '''
    global VERBOSE
    VERBOSE = int(value)


def dbg(*args, level=VERBOSE_NORMAL):
    '''Print a verbose debug logging message.

    :param args: sequence of arguments to print.
    :param value: verbosity level to set, e.g. VERBOSE_VERY.

    The message is only printed if level is at least the current
    verbosity level.'''
    if level > VERBOSE:
        return
    print(*args)


def inf(*args, colorize=False):
    '''Print an informational message.

    :param args: sequence of arguments to print.
    :param colorize: If this is True, the configuration option ``color.ui``
                     is undefined or true, and stdout is a terminal, then
                     the message is printed in green.
    '''

    if not _use_colors():
        colorize = False

    # This approach colorizes any sep= and end= text too, as expected.
    #
    # colorama automatically strips the ANSI escapes when stdout isn't a
    # terminal (by wrapping sys.stdout).
    if colorize:
        print(colorama.Fore.LIGHTGREEN_EX, end='')

    print(*args)

    if colorize:
        _reset_colors(sys.stdout)


def banner(*args):
    '''Prints args as a "banner" at inf() level.

    The args are prefixed with '=== ' and colorized by default.'''
    inf('===', *args, colorize=True)


def small_banner(*args):
    '''Prints args as a smaller banner(), i.e. prefixed with '-- ' and
    not colorized.'''
    inf('---', *args, colorize=False)


def wrn(*args):
    '''Print a warning.

    :param args: sequence of arguments to print.

    The message is prefixed with the string ``"WARNING: "``.

    If the configuration option ``color.ui`` is undefined or true and
    stdout is a terminal, then the message is printed in yellow.'''

    if _use_colors():
        print(colorama.Fore.LIGHTYELLOW_EX, end='', file=sys.stderr)

    print('WARNING: ', end='', file=sys.stderr)
    print(*args, file=sys.stderr)

    if _use_colors():
        _reset_colors(sys.stderr)


def err(*args, fatal=False):
    '''Print an error.

    This function does not abort the program. For that, use `die()`.

    :param args: sequence of arguments to print.
    :param fatal: if True, the the message is prefixed with "FATAL ERROR: ";
                  otherwise, "ERROR: " is used.

    If the configuration option ``color.ui`` is undefined or true and
    stdout is a terminal, then the message is printed in red.'''

    if _use_colors():
        print(colorama.Fore.LIGHTRED_EX, end='', file=sys.stderr)

    print('FATAL ERROR: ' if fatal else 'ERROR: ', end='', file=sys.stderr)
    print(*args, file=sys.stderr)

    if _use_colors():
        _reset_colors(sys.stderr)


def die(*args, exit_code=1):
    '''Print a fatal error, and abort the program.

    :param args: sequence of arguments to print.
    :param exit_code: return code the program should use when aborting.

    Equivalent to ``die(*args, fatal=True)``, followed by an attempt to
    abort with the given *exit_code*.'''
    err(*args, fatal=True)
    sys.exit(exit_code)


_COLOR_UI_WARNED = False

def _use_colors():
    # Convenience function for reading the color.ui setting
    try:
        return config.config.getboolean('color', 'ui', fallback=True)
    except ValueError as e:
        global _COLOR_UI_WARNED
        if not _COLOR_UI_WARNED:
            print("WARNING: invalid color.ui value: {}.".format(e),
                  file=sys.stderr)
            print('         To fix, run: "west config color.ui <true|false>"',
                  file=sys.stderr)
            _COLOR_UI_WARNED = True
        return False


def _reset_colors(file):
    # The flush=True avoids issues with unrelated output from commands (usually
    # Git) becoming colorized, due to the final attribute reset ANSI escape
    # getting line-buffered
    print(colorama.Style.RESET_ALL, end='', file=file, flush=True)



# Copyright 2019 Foundries.io Limited.
#
# SPDX-License-Identifier: Apache-2.0

'''Backports for older Python versions.'''

# NOTE: these should be periodically cleaned up and removed from the
# code base when support for older Python versions is dropped.

import sys

# Ensure CompletedProcess is available (it was added in 3.5).
if sys.version_info < (3, 5):
    from subprocess import CalledProcessError

    class CompletedProcess:
        '''subprocess.CompletedProcess-alike provided for Python 3.4.'''

        def __init__(self, args, returncode, stdout, stderr):
            self.args = args
            self.returncode = returncode
            self.stdout = stdout
            self.stderr = stderr

        def check_returncode(self):
            if self.returncode:
                # The stdout and stderr attributes weren't added until
                # Python 3.5; we manually include them here for code
                # targeting more recent versions which expects them.
                ret = CalledProcessError(self.returncode, self.args,
                                         output=self.stdout)
                ret.stdout = self.stdout
                ret.stderr = None
                raise ret
else:
    from subprocess import CompletedProcess

__all__ = ['CompletedProcess']



# Copyright (c) 2018, 2019 Nordic Semiconductor ASA
# Copyright 2018, 2019 Foundries.io
#
# SPDX-License-Identifier: Apache-2.0

'''West project commands'''

import argparse
import collections
from functools import partial
import os
from os.path import join, relpath, basename, dirname, exists, isdir
import shutil
import subprocess
import sys
import textwrap
import yaml

from west.configuration import config, update_config
from west import log
from west import util
from west.commands import WestCommand, CommandError
from west.manifest import Manifest, MalformedManifest, MalformedConfig, \
    MANIFEST_PROJECT_INDEX, ManifestProject
from west.manifest import MANIFEST_REV_BRANCH as MANIFEST_REV
from west.manifest import QUAL_MANIFEST_REV_BRANCH as QUAL_MANIFEST_REV
from urllib.parse import urlparse
import posixpath

#
# Project-related or multi-repo commands, like "init", "update",
# "diff", etc.
#

class _ProjectCommand(WestCommand):
    # Helper class which contains common code needed by various commands
    # in this file.

    def _parser(self, parser_adder, **kwargs):
        # Create and return a "standard" parser.

        kwargs['help'] = self.help
        kwargs['description'] = self.description
        kwargs['formatter_class'] = argparse.RawDescriptionHelpFormatter
        return parser_adder.add_parser(self.name, **kwargs)

    def _add_projects_arg(self, parser):
        # Adds a "projects" argument to the given parser.

        parser.add_argument('projects', metavar='PROJECT', nargs='*',
                            help='''projects (by name or path) to operate on;
                            defaults to all projects in the manifest''')

    def _cloned_projects(self, args):
        # Returns _projects(args.projects, only_cloned=True) if
        # args.projects is not empty (i.e., explicitly given projects
        # are required to be cloned). Otherwise, returns all cloned
        # projects.
        if args.projects:
            return self._projects(args.projects, only_cloned=True)
        else:
            return [p for p in self.manifest.projects if p.is_cloned()]

    def _projects(self, ids, only_cloned=False):
        try:
            return self.manifest.get_projects(ids, only_cloned=only_cloned)
        except ValueError as ve:
            if len(ve.args) != 2:
                raise          # not directly raised by get_projects()

            # Die with an error message on unknown or uncloned projects.
            unknown, uncloned = ve.args
            if unknown:
                log.die('Unknown project name{s}/path{s}: {unknown} '
                        '(use "west list" to list all projects)'
                        .format(s='s' if len(unknown) > 1 else '',
                                unknown=', '.join(unknown)))
            elif only_cloned and uncloned:
                plural = len(uncloned) > 1
                names = [p.name for p in uncloned]
                log.die(textwrap.dedent('''\
                The following project{} not cloned: {}.
                Please clone {} first, with:
                    west update {}
                then retry.'''.format('s are' if plural else ' is',
                                      ", ".join(names),
                                      'them' if plural else 'it',
                                      " ".join(names))))
            else:
                # Should never happen, but re-raise to fail fast and
                # preserve a stack trace, to encourage a bug report.
                raise

    def _handle_failed(self, args, failed):
        # Shared code for commands (like status, diff, update) that need
        # to do the same thing to multiple projects, but collect
        # and report errors if anything failed.

        if not failed:
            return
        elif len(failed) < 20:
            log.err('{command} failed for project{s} {projects}'.
                    format(command=self.name,
                           s='s:' if len(failed) > 1 else '',
                           projects=', '.join(p.format('{name}')
                                              for p in failed)))
        else:
            log.err('{command} failed for multiple projects; see above')
        raise CommandError(1)

class Init(_ProjectCommand):

    def __init__(self):
        super().__init__(
            'init',
            'create a west installation',
            textwrap.dedent('''\
            Creates a west installation as follows:

              1. Creates a .west directory and clones a manifest repository
                 from a git URL to a temporary subdirectory of .west,
                 .west/<tmpdir>.
              2. Parses the manifest file, .west/<tmpdir>/west.yml.
                 This file's contents can specify manifest.path, the location
                 of the manifest repository in the installation, like so:

                 manifest:
                    self:
                      path: <manifest.path value>

                 If left unspecified, the basename of the manifest
                 repository URL is used as manifest.path (so for example,
                 "http://.../foo/bar" results in "bar").
              3. Creates a local west configuration file, .west/config,
                 and sets the manifest.path option there to the value
                 from step 2. (Run "west config -h" for details on west
                 configuration files.)
              4. Moves the manifest repository from .west/<tmpdir> to
                 manifest.path, next to the .west directory.

            The default manifest repository URL is:

            {}

            This can be overridden using -m.

            The default revision in this repository to check out is "{}";
            override with --mr.'''.format(MANIFEST_URL_DEFAULT,
                                          MANIFEST_REV_DEFAULT)),
            requires_installation=False)

    def do_add_parser(self, parser_adder):
        parser = self._parser(parser_adder)

        parser.add_argument('-m', '--manifest-url',
                            help='''manifest repository URL to clone;
                            cannot be combined with -l''')
        parser.add_argument('--mr', '--manifest-rev', dest='manifest_rev',
                            help='''manifest revision to check out and use;
                            cannot be combined with -l''')
        parser.add_argument('-l', '--local', action='store_true',
                            help='''use an existing local manifest repository
                            instead of cloning one; cannot be combined with
                            -m or --mr.''')

        parser.add_argument(
            'directory', nargs='?', default=None,
            help='''Directory to create the installation in (default: current
                 directory). Missing intermediate directories will be created.
                 If -l is given, this is the path to the manifest repository to
                 use instead.''')

        return parser

    def do_run(self, args, ignored):
        if self.topdir:
            zb = os.environ.get('ZEPHYR_BASE')
            if zb:
                msg = textwrap.dedent('''
                Note:
                    In your environment, ZEPHYR_BASE is set to:
                    {}

                    This forces west to search for an installation there.
                    Try unsetting ZEPHYR_BASE and re-running this command.'''.
                                      format(zb))
            else:
                msg = ''
            log.die('already initialized in {}, aborting.{}'.
                    format(self.topdir, msg))

        if args.local and (args.manifest_url or args.manifest_rev):
            log.die('-l cannot be combined with -m or --mr')

        if shutil.which('git') is None:
            log.die("can't find git; install it or ensure it's on your PATH")

        # west.manifest will try to read manifest.path and use it when
        # parsing the manifest. Clear it out for now so we can parse
        # the manifest without it; local() or bootstrap() will set it
        # properly.
        if config.get('manifest', 'path', fallback=None) is not None:
            config.remove_option('manifest', 'path')
        if args.local:
            projects, manifest_dir = self.local(args)
        else:
            projects, manifest_dir = self.bootstrap(args)

        self.fixup_zephyr_base(projects)

        log.banner('Initialized. Now run "west update" inside {}.'.
                   format(self.topdir))

    def local(self, args):
        if args.manifest_rev is not None:
            log.die('--mr cannot be used with -l')

        manifest_dir = util.canon_path(args.directory or os.getcwd())
        manifest_file = join(manifest_dir, 'west.yml')
        topdir = dirname(manifest_dir)
        rel_manifest = basename(manifest_dir)
        west_dir = os.path.join(topdir, WEST_DIR)

        log.banner('Initializing from existing manifest repository',
                   rel_manifest)
        if not exists(manifest_file):
            log.die('No "west.yml" found in {}'.format(manifest_dir))

        self.create(west_dir)
        os.chdir(topdir)
        # This validates the manifest. Note we cannot use
        # self.manifest from west init, as we are in the middle of
        # creating the installation right now.
        projects = self.projects(manifest_file)
        log.small_banner('Creating {} and local configuration'.
                         format(west_dir))
        update_config('manifest', 'path', rel_manifest)

        self.topdir = topdir

        return projects, manifest_dir

    def bootstrap(self, args):
        manifest_url = args.manifest_url or MANIFEST_URL_DEFAULT
        manifest_rev = args.manifest_rev or MANIFEST_REV_DEFAULT
        topdir = util.canon_path(args.directory or os.getcwd())
        west_dir = join(topdir, WEST_DIR)

        log.banner('Initializing in', topdir)
        if not isdir(topdir):
            self.create(topdir, exist_ok=False)
        os.chdir(topdir)

        # Clone the manifest repository into a temporary directory. It's
        # important that west_dir exists and we're under topdir, or we
        # won't be able to call self.projects() without error later.
        tempdir = join(west_dir, 'manifest-tmp')
        if exists(tempdir):
            log.dbg('removing existing temporary manifest directory', tempdir)
            shutil.rmtree(tempdir)
        try:
            self.clone_manifest(manifest_url, manifest_rev, tempdir)
            temp_manifest_file = join(tempdir, 'west.yml')
            if not exists(temp_manifest_file):
                log.die('No "west.yml" in manifest repository ({})'.
                        format(tempdir))

            projects = self.projects(temp_manifest_file)
            manifest_project = projects[MANIFEST_PROJECT_INDEX]
            if manifest_project.path:
                manifest_path = manifest_project.path
                manifest_abspath = join(topdir, manifest_path)
            else:
                url_path = urlparse(manifest_url).path
                manifest_path = posixpath.basename(url_path)
                manifest_abspath = join(topdir, manifest_path)

            shutil.move(tempdir, manifest_abspath)
            update_config('manifest', 'path', manifest_path)
        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

        self.topdir = topdir

        return projects, manifest_project.abspath

    def projects(self, manifest_file):
        try:
            return Manifest.from_file(manifest_file).projects
        except MalformedManifest as mm:
            log.die(mm.args[0])
        except MalformedConfig as mc:
            log.die(mc.args[0])

    def fixup_zephyr_base(self, projects):
        for project in projects:
            if project.path == 'zephyr':
                update_config('zephyr', 'base', project.path)

    def create(self, directory, exist_ok=True):
        try:
            os.makedirs(directory, exist_ok=exist_ok)
        except PermissionError:
            log.die('Cannot initialize in {}: permission denied'.
                    format(directory))
        except FileExistsError:
            log.die('Something else created {} concurrently; quitting'.
                    format(directory))
        except Exception as e:
            log.die("Can't create directory {}: {}".format(directory, e.args))

    def clone_manifest(self, url, rev, dest, exist_ok=False):
        log.small_banner('Cloning manifest repository from {}, rev. {}'.
                         format(url, rev))
        if not exist_ok and exists(dest):
            log.die('refusing to clone into existing location ' + dest)

        subprocess.check_call(('git', 'init', dest))
        subprocess.check_call(('git', 'remote', 'add', 'origin', '--', url),
                              cwd=dest)
        maybe_sha = _maybe_sha(rev)
        if maybe_sha:
            # Fetch the ref-space and hope the SHA is contained in
            # that ref-space
            subprocess.check_call(('git', 'fetch', 'origin', '--tags',
                                   '--', 'refs/heads/*:refs/remotes/origin/*'),
                                  cwd=dest)
        else:
            # Fetch the ref-space similar to git clone plus the ref
            # given by user.  Redundancy is ok, for example if the user
            # specifies 'heads/master'. This allows users to specify:
            # pull/<no>/head for pull requests
            subprocess.check_call(('git', 'fetch', 'origin', '--tags', '--',
                                   rev, 'refs/heads/*:refs/remotes/origin/*'),
                                  cwd=dest)

        try:
            # Using show-ref to determine if rev is available in local repo.
            subprocess.check_call(('git', 'show-ref', '--', rev), cwd=dest)
            local_rev = True
        except subprocess.CalledProcessError:
            local_rev = False

        if local_rev or maybe_sha:
            subprocess.check_call(('git', 'checkout', rev), cwd=dest)
        else:
            subprocess.check_call(('git', 'checkout', 'FETCH_HEAD'), cwd=dest)


class List(_ProjectCommand):
    def __init__(self):
        super().__init__(
            'list',
            'print information about projects in the west manifest',
            textwrap.dedent('''\
            Print information about projects in the west manifest,
            using format strings.'''))

    def do_add_parser(self, parser_adder):
        default_fmt = '{name:12} {path:28} {revision:40} {url}'
        parser = self._parser(
            parser_adder,
            epilog=textwrap.dedent('''\
            FORMAT STRINGS
            --------------

            Projects are listed using a Python 3 format string. Arguments
            to the format string are accessed by name.

            The default format string is:

            "{}"

            The following arguments are available:

            - name: project name in the manifest
            - url: full remote URL as specified by the manifest
            - path: the relative path to the project from the top level,
              as specified in the manifest where applicable
            - abspath: absolute and normalized path to the project
            - posixpath: like abspath, but in posix style, that is, with '/'
              as the separator character instead of '\\'
            - revision: project's revision as it appears in the manifest
            - sha: project's revision as a SHA. Note that use of this requires
              that the project has been cloned.
            - cloned: "cloned" if the project has been cloned, "not-cloned"
              otherwise
            - clone_depth: project clone depth if specified, "None" otherwise
            '''.format(default_fmt)))
        parser.add_argument('-a', '--all', action='store_true',
                            help='ignored for backwards compatibility'),
        parser.add_argument('-f', '--format', default=default_fmt,
                            help='''format string to use to list each
                            project; see FORMAT STRINGS below.''')

        self._add_projects_arg(parser)

        return parser

    def do_run(self, args, user_args):
        def sha_thunk(project):
            if not project.is_cloned():
                log.die('cannot get sha for uncloned project {0}; '
                        'run "west update {0}" and retry'.
                        format(project.name))
            elif project.revision:
                return project.sha(MANIFEST_REV)
            else:
                return '{:40}'.format('N/A')

        def cloned_thunk(project):
            return "cloned" if project.is_cloned() else "not-cloned"

        def delay(func, project):
            return DelayFormat(partial(func, project))

        for project in self._projects(args.projects):
            # Spelling out the format keys explicitly here gives us
            # future-proofing if the internal Project representation
            # ever changes.
            #
            # Using DelayFormat delays computing derived values, such
            # as SHAs, unless they are specifically requested, and then
            # ensures they are only computed once.
            try:
                result = args.format.format(
                    name=project.name,
                    url=project.url or 'N/A',
                    path=project.path,
                    abspath=project.abspath,
                    posixpath=project.posixpath,
                    revision=project.revision or 'N/A',
                    clone_depth=project.clone_depth or "None",
                    cloned=delay(cloned_thunk, project),
                    sha=delay(sha_thunk, project))
            except KeyError as e:
                # The raised KeyError seems to just put the first
                # invalid argument in the args tuple, regardless of
                # how many unrecognizable keys there were.
                log.die('unknown key "{}" in format string "{}"'.
                        format(e.args[0], args.format))
            except IndexError:
                self.parser.print_usage()
                log.die('invalid format string', args.format)
            except subprocess.CalledProcessError:
                log.die('subprocess error while listing project',
                        project.name)

            log.inf(result, colorize=False)


class ManifestCommand(_ProjectCommand):
    # The slightly weird naming is to avoid a conflict with
    # west.manifest.Manifest.

    def __init__(self):
        super(ManifestCommand, self).__init__(
            'manifest',
            'manage the west manifest',
            textwrap.dedent('''\
            Manages the west manifest.

            The --freeze operation outputs the manifest with all
            project-related values fully specified: defaults are
            applied, and all revisions are converted to SHAs based on
            the current manifest-rev branches. Note that all projects
            must be cloned for this to work.

            The --validate operation validates the current manifest,
            printing an error if it cannot be successfully parsed.'''),
            accepts_unknown_args=False)

    def do_add_parser(self, parser_adder):
        parser = self._parser(parser_adder)

        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('--freeze', action='store_true',
                           help='emit a manifest with SHAs for each revision')
        group.add_argument('--validate', action='store_true',
                           help='''validate the current manifest,
                           exiting with an error if there are issues''')

        group = parser.add_argument_group('options for --freeze')
        group.add_argument('-o', '--out',
                           help='output file, default is standard output')

        return parser

    def do_run(self, args, user_args):
        if args.freeze:
            self._freeze(args)
        else:
            # --validate. The exception block in main() handles errors.
            Manifest.from_file()

    def _freeze(self, args):
        try:
            frozen = Manifest.from_file().as_frozen_dict()
        except RuntimeError as re:
            log.die(*(list(re.args) + ['(run "west update" and retry)']))

        # This is a destructive operation, so it's done here to avoid
        # impacting code which doesn't expect this representer to be
        # in place.
        yaml.SafeDumper.add_representer(collections.OrderedDict, self._rep)

        if args.out:
            with open(args.out, 'w') as f:
                yaml.safe_dump(frozen, default_flow_style=False, stream=f)
        else:
            yaml.safe_dump(frozen, default_flow_style=False, stream=sys.stdout)

    def _rep(self, dumper, value):
        # See https://yaml.org/type/map.html for details on the tag.
        return util._represent_ordered_dict(dumper, 'tag:yaml.org,2002:map',
                                            value)

class Diff(_ProjectCommand):
    def __init__(self):
        super().__init__(
            'diff',
            '"git diff" for one or more projects',
            'Runs "git diff" on each of the specified projects.')

    def do_add_parser(self, parser_adder):
        parser = self._parser(parser_adder)
        self._add_projects_arg(parser)
        return parser

    def do_run(self, args, ignored):
        failed = []
        for project in self._cloned_projects(args):
            log.banner(project.format('diff for {name_and_path}:'))
            # Use paths that are relative to the base directory to make it
            # easier to see where the changes are
            try:
                project.git('diff --src-prefix={path}/ --dst-prefix={path}/')
            except subprocess.CalledProcessError:
                failed.append(project)
        self._handle_failed(args, failed)


class Status(_ProjectCommand):
    def __init__(self):
        super().__init__(
            'status',
            '"git status" for one or more projects',
            "Runs 'git status' for each of the specified projects.")

    def do_add_parser(self, parser_adder):
        parser = self._parser(parser_adder)
        self._add_projects_arg(parser)
        return parser

    def do_run(self, args, user_args):
        failed = []
        for project in self._cloned_projects(args):
            log.banner(project.format('status of {name_and_path}:'))
            try:
                project.git('status', extra_args=user_args)
            except subprocess.CalledProcessError:
                failed.append(project)
        self._handle_failed(args, failed)


class Update(_ProjectCommand):

    def __init__(self):
        super().__init__(
            'update',
            'update projects described in west.yml',
            textwrap.dedent('''\
            Updates each project repository to the revision specified in
            the manifest file, west.yml, as follows:

              1. Fetch the project's remote to ensure the manifest
                 revision is available locally
              2. Reset the manifest-rev branch to the revision in
                 west.yml
              3. Check out the new manifest-rev commit as a detached HEAD
                 (but see "checked out branches")

            You must have already created a west installation with "west init".

            This command does not alter the manifest repository's contents.''')
        )

    def do_add_parser(self, parser_adder):
        parser = self._parser(parser_adder)

        group = parser.add_argument_group(
            title='fetching behavior',
            description='By default, west update tries to avoid fetching.')
        group.add_argument('-f', '--fetch', dest='fetch_strategy',
                           choices=['always', 'smart'],
                           help='''how to fetch projects when updating:
                           "always" fetches every project before update,
                           while "smart" (default) skips fetching projects
                           whose revisions are SHAs or tags available
                           locally''')

        group = parser.add_argument_group(
            title='checked out branch behavior',
            description=textwrap.dedent('''\
            By default, locally checked out branches are left behind
            when manifest-rev commits are checked out.'''))
        group.add_argument('-k', '--keep-descendants', action='store_true',
                           help='''if a checked out branch is a descendant
                           of the new manifest-rev, leave it checked out
                           instead (takes priority over --rebase)''')
        group.add_argument('-r', '--rebase', action='store_true',
                           help='''rebase any checked out branch onto the new
                           manifest-rev instead (leaving behind partial
                           rebases on error)''')

        group = parser.add_argument_group('deprecated options')
        group.add_argument('-x', '--exclude-west', action='store_true',
                           help='ignored for backwards compatibility')

        self._add_projects_arg(parser)

        return parser

    def do_run(self, args, user_args):
        if args.exclude_west:
            log.wrn('ignoring --exclude-west')
        fs = self._fetch_strategy(args)

        failed = []
        for project in self._projects(args.projects):
            if isinstance(project, ManifestProject):
                continue
            log.banner(project.format('updating {name_and_path}:'))
            try:
                _update(project, fs, args.rebase, args.keep_descendants)
            except subprocess.CalledProcessError:
                failed.append(project)
        self._handle_failed(args, failed)

    def _fetch_strategy(self, args):
        cfg = config.get('update', 'fetch', fallback=None)
        if cfg is not None and cfg not in ('always', 'smart'):
            log.wrn('ignoring invalid config update.fetch={}; '
                    'choices: always, smart'.format(cfg))
            cfg = None
        if args.fetch_strategy:
            return args.fetch_strategy
        elif cfg:
            return cfg
        else:
            return 'smart'


class ForAll(_ProjectCommand):
    def __init__(self):
        super().__init__(
            'forall',
            'run a command in one or more local projects',
            textwrap.dedent('''\
            Runs a shell (on a Unix OS) or batch (on Windows) command
            within the repository of each of the specified PROJECTs.

            If the command has multiple words, you must quote the -c
            option to prevent the shell from splitting it up. Since
            the command is run through the shell, you can use
            wildcards and the like.

            For example, the following command will list the contents
            of proj-1's and proj-2's repositories on Linux and macOS,
            in long form:

                west forall -c "ls -l" proj-1 proj-2
            '''))

    def do_add_parser(self, parser_adder):
        parser = self._parser(parser_adder)
        parser.add_argument('-c', dest='command', metavar='COMMAND',
                            required=True)
        self._add_projects_arg(parser)
        return parser

    def do_run(self, args, user_args):
        failed = []
        for project in self._cloned_projects(args):
            log.banner(project.format('running "{c}" in {name_and_path}:',
                                      c=args.command))
            rc = subprocess.Popen(args.command, shell=True,
                                  cwd=project.abspath).wait()
            if rc:
                failed.append(project)
        self._handle_failed(args, failed)


class SelfUpdate(_ProjectCommand):
    def __init__(self):
        super().__init__(
            'selfupdate',
            'deprecated; exists for backwards compatibility',
            'Do not use. You can upgrade west with pip only from v0.6.0.')

    def do_add_parser(self, parser_adder):
        return self._parser(parser_adder)

    def do_run(self, args, user_args):
        log.die(self.description)

#
# Private helper routines.
#

def _maybe_sha(rev):
    # Return true if and only if the given revision might be a SHA.

    try:
        int(rev, 16)
    except ValueError:
        return False

    return len(rev) <= 40


def _update(project, fetch, rebase, keep_descendants):
    if not project.is_cloned():
        _clone(project)

    if fetch == 'always' or _rev_type(project) not in ('tag', 'commit'):
        _fetch(project)
    else:
        log.dbg('skipping unnecessary fetch')
        project.git('update-ref ' + QUAL_MANIFEST_REV +
                    ' {revision}^{{commit}}')

    try:
        sha = project.sha(QUAL_MANIFEST_REV)
    except subprocess.CalledProcessError:
        # This is a sign something's really wrong. Add more help.
        log.err(project.format("no SHA for branch {mr} in {name_and_path}; "
                               'was the branch deleted?', mr=MANIFEST_REV))
        raise

    cp = project.git('rev-parse --abbrev-ref HEAD', capture_stdout=True)
    current_branch = cp.stdout.decode('utf-8').strip()
    if current_branch != 'HEAD':
        is_ancestor = project.is_ancestor_of(sha, current_branch)
        try_rebase = rebase
    else:  # HEAD means no branch is checked out.
        # If no branch is checked out, 'rebase' and 'keep_descendants' don't
        # matter.
        is_ancestor = False
        try_rebase = False

    if keep_descendants and is_ancestor:
        # A descendant is currently checked out and keep_descendants was
        # given, so there's nothing more to do.
        log.small_banner(project.format(
            '{name}: left descendant branch "{b}" checked out',
            b=current_branch))
    elif try_rebase:
        # Attempt a rebase.
        log.small_banner(project.format('{name}: rebasing to ' + MANIFEST_REV))
        project.git('rebase ' + QUAL_MANIFEST_REV)
    else:
        # We can't keep a descendant or rebase, so just check
        # out the new detached HEAD, then print some helpful context.
        project.git('checkout --detach --quiet ' + sha)
        log.small_banner(project.format(
            "{name}: checked out {r} as detached HEAD", r=sha))
        _post_checkout_help(project, current_branch, sha, is_ancestor)


def _clone(project):
    log.small_banner(project.format('{name}: cloning and initializing'))
    project.git('init {abspath}', cwd=util.west_topdir())
    # This remote is only added for the user's convenience. We always fetch
    # directly from the URL specified in the manifest.
    project.git('remote add -- {remote_name} {url}')


def _rev_type(project):
    # Returns a "refined" revision type of project.revision as of the
    # following strings: 'tag', 'tree', 'blob', 'commit', 'branch',
    # 'other'.
    #
    # The approach combines git cat-file -t and git rev-parse because,
    # while cat-file can for sure tell us a blob, tree, or tag, it
    # doesn't have a way to disambiguate between branch names and
    # other types of commit-ishes, like SHAs, things like "HEAD" or
    # "HEAD~2", etc.
    #
    # We need this extra layer of refinement to be able to avoid
    # fetching SHAs that are already available locally.
    #
    # This doesn't belong in manifest.py because it contains "west
    # update" specific logic.

    cp = project.git('cat-file -t {revision}', check=False,
                     capture_stdout=True, capture_stderr=True)
    stdout = cp.stdout.decode('utf-8').strip()
    if cp.returncode:
        return 'other'
    elif stdout in ('blob', 'tree', 'tag'):
        return stdout
    elif stdout != 'commit':    # just future-proofing
        return 'other'

    # to tell branches apart from commits, we need rev-parse.
    cp = project.git('rev-parse --verify --symbolic-full-name {revision}',
                     check=False, capture_stdout=True, capture_stderr=True)
    if cp.returncode:
        # This can happen if the ref name is ambiguous, e.g.:
        #
        # $ git update-ref ambiguous-ref HEAD~2
        # $ git checkout -B ambiguous-ref
        #
        # Which creates both .git/ambiguous-ref and
        # .git/refs/heads/ambiguous-ref.
        return 'other'

    stdout = cp.stdout.decode('utf-8').strip()
    if stdout.startswith('refs/heads'):
        return 'branch'
    elif not stdout:
        return 'commit'
    else:
        return 'other'


def _fetch(project):
    # Fetches upstream changes for 'project' and updates the 'manifest-rev'
    # branch to point to the revision specified in the manifest.

    # Fetch the revision specified in the manifest into the local ref space.
    #
    # The following two-step approach avoids a "trying to write
    # non-commit object" error when the revision is an annotated
    # tag. ^{commit} type peeling isn't supported for the <src> in a
    # <src>:<dst> refspec, so we have to do it separately.
    msg = "{name}: fetching, need revision {revision}"
    if project.clone_depth:
        msg += " with --depth {clone_depth}"
    # -f is needed to avoid errors in case multiple remotes are present,
    # at least one of which contains refs that can't be fast-forwarded to our
    # local ref space.
    #
    # --tags is required to get tags when the remote is specified as a URL.
    fetch_cmd = ('fetch -f --tags' +
                 ('--depth={clone_depth} ' if project.clone_depth else ' ') +
                 '-- {url} ')
    update_cmd = 'update-ref ' + QUAL_MANIFEST_REV + ' '
    if _maybe_sha(project.revision):
        # Don't fetch a SHA directly, as server may restrict from doing so.
        fetch_cmd += 'refs/heads/*:refs/west/*'
        update_cmd += '{revision}'
    else:
        # The revision is definitely not a SHA, so it's safe to fetch directly.
        # This avoids fetching unnecessary ref space from the remote.
        fetch_cmd += '{revision}'
        update_cmd += 'FETCH_HEAD^{{commit}}'

    log.small_banner(project.format(msg))
    project.git(fetch_cmd)
    project.git(update_cmd)

    if not _head_ok(project):
        # If nothing it checked out (which would usually only happen just after
        # we initialize the repository), check out 'manifest-rev' in a detached
        # HEAD state.
        #
        # Otherwise, the initial state would have nothing checked out, and HEAD
        # would point to a non-existent refs/heads/master branch (that would
        # get created if the user makes an initial commit). That state causes
        # e.g. 'west rebase' to fail, and might look confusing.
        #
        # The --detach flag is strictly redundant here, because the
        # refs/heads/<branch> form already detaches HEAD, but it avoids a
        # spammy detached HEAD warning from Git.
        project.git('checkout --detach ' + QUAL_MANIFEST_REV)


def _head_ok(project):
    # Returns True if the reference 'HEAD' exists and is not a tag or remote
    # ref (e.g. refs/remotes/origin/HEAD).
    # Some versions of git will report 1, when doing
    # 'git show-ref --verify HEAD' even if HEAD is valid, see #119.
    # 'git show-ref --head <reference>' will always return 0 if HEAD or
    # <reference> is valid.
    # We are only interested in HEAD, thus we must avoid <reference> being
    # valid. '/' can never point to valid reference, thus 'show-ref --head /'
    # will return:
    # - 0 if HEAD is present
    # - 1 otherwise
    return project.git('show-ref --quiet --head /',
                       check=False).returncode == 0


def _post_checkout_help(project, branch, sha, is_ancestor):
    # Print helpful information to the user about a project that
    # might have just left a branch behind.

    if branch == 'HEAD':
        # If there was no branch checked out, there are no
        # additional diagnostics that need emitting.
        return

    rel = relpath(project.abspath)
    if is_ancestor:
        # If the branch we just left behind is a descendant of
        # the new HEAD (e.g. if this is a topic branch the
        # user is working on and the remote hasn't changed),
        # print a message that makes it easy to get back,
        # no matter where in the installation os.getcwd() is.
        log.wrn(project.format(
            'left behind {name} branch "{b}"; to switch '
            'back to it (fast forward), use: git -C {rp} checkout {b}',
            b=branch, rp=rel))
        log.dbg('(To do this automatically in the future,',
                'use "west update --keep-descendants".)')
    else:
        # Tell the user how they could rebase by hand, and
        # point them at west update --rebase.
        log.wrn(project.format(
            'left behind {name} branch "{b}"; '
            'to rebase onto the new HEAD: git -C {rp} rebase {sh} {b}',
            b=branch, rp=rel, sh=sha))
        log.dbg('(To do this automatically in the future,',
                'use "west update --rebase".)')


#
# Special files and directories in the west installation.
#
# These are given variable names for clarity, but they can't be
# changed without propagating the changes into west itself.
#

# Top-level west directory, containing west itself and the manifest.
WEST_DIR = '.west'

# Manifest repository directory under WEST_DIR.
MANIFEST = 'manifest'
# Default manifest repository URL.
MANIFEST_URL_DEFAULT = 'https://github.com/zephyrproject-rtos/zephyr'
# Default revision to check out of the manifest repository.
MANIFEST_REV_DEFAULT = 'master'

#
# Helper class for creating format string keys that are expensive or
# undesirable to compute if not needed.
#

class DelayFormat:
    '''Delays formatting an object.'''

    def __init__(self, obj):
        '''Delay formatting `obj` until a format operation using it.

        :param obj: object to format

        If callable(obj) returns True, then obj() will be used as the
        string to be formatted. Otherwise, str(obj) is used.'''
        self.obj = obj
        self.as_str = None

    def __format__(self, format_spec):
        if self.as_str is None:
            if callable(self.obj):
                self.as_str = self.obj()
                assert isinstance(self.as_str, str)
            else:
                self.as_str = str(self.obj)
        return ('{:' + format_spec + '}').format(self.as_str)



# Copyright 2018 Open Source Foundries Limited.
#
# SPDX-License-Identifier: Apache-2.0

# Some code (the PyYAML representer for an OrderedDict) is adapted
# from the PyYAML source code, which is MIT Licensed:
#
# SPDX-License-Identifier: MIT
#
# Full license text from the PyYAML repository is reproduced below:
#
# Copyright (c) 2017-2018 Ingy döt Net
# Copyright (c) 2006-2016 Kirill Simonov
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''Miscellaneous utilities.
'''

import os
import pathlib
import shlex
import textwrap

import yaml

def canon_path(path):
    '''Returns a canonical version of the path.

    This is currently ``os.path.normcase(os.path.abspath(path))``. The
    path separator is converted to os.sep on platforms where that
    matters (Windows).

    :param path: path whose canonical name to return; need not
                 refer to an existing file.
    '''
    return os.path.normcase(os.path.abspath(path))

def escapes_directory(path, directory):
    '''Returns True if `path` escapes parent directory `directory`.

    :param path: path to check is inside `directory`
    :param directory: parent directory to check

    Verifies `path` is inside of `directory`, after computing
    normalized real paths for both.'''
    # It turns out not to be easy to implement this without using
    # pathlib.
    p = pathlib.Path(os.path.normcase(os.path.realpath(path)))
    d = pathlib.Path(os.path.normcase(os.path.realpath(directory)))
    try:
        p.relative_to(d)
        ret = False
    except ValueError:
        ret = True
    return ret


def quote_sh_list(cmd):
    '''Transform a command from list into shell string form.'''
    fmt = ' '.join('{}' for _ in cmd)
    args = [shlex.quote(s) for s in cmd]
    return fmt.format(*args)


def wrap(text, indent):
    '''Convenience routine for wrapping text to a consistent indent.'''
    return textwrap.wrap(text, initial_indent=indent,
                         subsequent_indent=indent)


class WestNotFound(RuntimeError):
    '''Neither the current directory nor any parent has a West installation.'''


def west_dir(start=None):
    '''Returns the absolute path of the installation's .west directory.

    Starts the search from the start directory, and goes to its
    parents. If the start directory is not specified, the current
    directory is used.

    Raises WestNotFound if no .west directory is found.
    '''
    return os.path.join(west_topdir(start), '.west')


def west_topdir(start=None, fall_back=True):
    '''
    Like west_dir(), but returns the path to the parent directory of the .west/
    directory instead, where project repositories are stored
    '''
    # If you change this function, make sure to update the bootstrap
    # script's find_west_topdir().

    cur_dir = start or os.getcwd()

    while True:
        if os.path.isdir(os.path.join(cur_dir, '.west')):
            return cur_dir

        parent_dir = os.path.dirname(cur_dir)
        if cur_dir == parent_dir:
            # At the root. Should we fall back?
            if fall_back and os.environ.get('ZEPHYR_BASE'):
                return west_topdir(os.environ['ZEPHYR_BASE'],
                                   fall_back=False)
            else:
                raise WestNotFound('Could not find a West installation '
                                   'in this or any parent directory')
        cur_dir = parent_dir


def _represent_ordered_dict(dumper, tag, mapping, flow_style=None):
    # PyYAML representer for ordered dicts. Used internally.

    value = []
    node = yaml.MappingNode(tag, value, flow_style=flow_style)
    if dumper.alias_key is not None:
        dumper.represented_objects[dumper.alias_key] = node
    best_style = True
    if hasattr(mapping, 'items'):
        mapping = list(mapping.items())
        # The only real difference between
        # BaseRepresenter.represent_mapping and this function is that
        # we omit the sort here. Ugh!
    for item_key, item_value in mapping:
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)
        if not (isinstance(node_key, yaml.ScalarNode) and
                not node_key.style):
            best_style = False
        if not (isinstance(node_value, yaml.ScalarNode) and
                not node_value.style):
            best_style = False
        value.append((node_key, node_value))
    if flow_style is None:
        if dumper.default_flow_style is not None:
            node.flow_style = dumper.default_flow_style
        else:
            node.flow_style = best_style
    return node



# Copyright (c) 2019, Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0

'''West config commands'''

import argparse
import configparser

from west import log
from west.configuration import read_config, update_config, delete_config, \
    ConfigFile
from west.commands import WestCommand, CommandError

CONFIG_DESCRIPTION = '''\
West configuration file handling.

West follows Git-like conventions for configuration file locations.
There are three types of configuration file: system-wide files apply
to all users on the current machine, global files apply to the current
user, and local files apply to the current west installation.

System files:

- Linux: /etc/westconfig
- macOS: /usr/local/etc/westconfig
- Windows: %PROGRAMDATA%\\west\\config

Global files:

- Linux: ~/.westconfig or (if $XDG_CONFIG_HOME is set)
  $XDG_CONFIG_HOME/west/config
- macOS: ~/.westconfig
- Windows: .westconfig in the user's home directory, as determined
  by os.path.expanduser.

Local files:

- Linux, macOS, Windows: <installation-root-directory>/.west/config

You can override these files' locations with the WEST_CONFIG_SYSTEM,
WEST_CONFIG_GLOBAL, and WEST_CONFIG_LOCAL environment variables.

Configuration values from later configuration files override configuration
from earlier ones. Local values have highest precedence, and system values
lowest.

To get a value for <name>, type:
    west config <name>

To set a value for <name>, type:
    west config <name> <value>

To list all options and their values:
    west config -l

To delete <name> in the local or global file (wherever it's set
first, not in both; if set locally, global values become visible):
    west config -d <name>

To delete <name> in the global file only:
    west config -d --global <name>

To delete <name> everywhere it's set, including the system file:
    west config -D <name>
'''

CONFIG_EPILOG = '''\
If the configuration file to use is not set, reads use all three in
precedence order, and writes use the local file.'''

ALL = ConfigFile.ALL
SYSTEM = ConfigFile.SYSTEM
GLOBAL = ConfigFile.GLOBAL
LOCAL = ConfigFile.LOCAL

class Once(argparse.Action):
    # For enforcing mutual exclusion of options by ensuring self.dest
    # can only be set once.
    #
    # This sets the 'configfile' attribute depending on the option string,
    # which must be --system, --global, or --local.

    def __call__(self, parser, namespace, ignored, option_string=None):
        values = {'--system': SYSTEM, '--global': GLOBAL, '--local': LOCAL}
        rev = {v: k for k, v in values.items()}

        if getattr(namespace, self.dest):
            previous = rev[getattr(namespace, self.dest)]
            parser.error("argument {}: not allowed with argument {}".
                         format(option_string, previous))

        setattr(namespace, self.dest, values[option_string])

class Config(WestCommand):
    def __init__(self):
        super().__init__(
            'config',
            'get or set configuration settings in west config files',
            CONFIG_DESCRIPTION)

    def do_add_parser(self, parser_adder):
        parser = parser_adder.add_parser(
            self.name,
            help=self.help,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=self.description,
            epilog=CONFIG_EPILOG)

        parser.add_argument('-l', '--list', action='store_true',
                            help='list all options and their values')
        parser.add_argument('-d', '--delete', action='store_true',
                            help='delete an option in one config file')
        parser.add_argument('-D', '--delete-all', action='store_true',
                            help="delete an option everywhere it's set")

        group = parser.add_argument_group(
            'configuration file to use (give at most one)')
        group.add_argument('--system', dest='configfile', nargs=0, action=Once,
                           help='system-wide file')
        group.add_argument('--global', dest='configfile', nargs=0, action=Once,
                           help='global (user-wide) file')
        group.add_argument('--local', dest='configfile', nargs=0, action=Once,
                           help="this installation's file")

        parser.add_argument('name', nargs='?',
                            help='''config option in section.key format;
                            e.g. "foo.bar" is section "foo", key "bar"''')
        parser.add_argument('value', nargs='?', help='value to set "name" to')

        return parser

    def do_run(self, args, user_args):
        delete = args.delete or args.delete_all
        if args.list:
            if args.name:
                self.parser.error('-l cannot be combined with name argument')
            elif delete:
                self.parser.error('-l cannot be combined with -d or -D')
        elif not args.name:
            self.parser.error('missing argument name '
                              '(to list all options and values, use -l)')
        elif args.delete and args.delete_all:
            self.parser.error('-d cannot be combined with -D')

        if args.list:
            self.list(args)
        elif delete:
            self.delete(args)
        elif args.value is None:
            self.read(args)
        else:
            self.write(args)

    def list(self, args):
        cfg = configparser.ConfigParser()
        what = args.configfile or ALL
        read_config(configfile=what, config=cfg)
        for s in cfg.sections():
            for k, v in cfg[s].items():
                log.inf('{}.{}={}'.format(s, k, v))

    def delete(self, args):
        section, key = self._sk(args)
        if args.delete_all:
            what = ALL
        elif args.configfile:
            what = args.configfile
        else:
            what = None   # local or global, whichever comes first

        try:
            delete_config(section, key, configfile=what)
        except KeyError:
            log.dbg('{} was not set in requested location(s)'.
                    format(args.name))
            raise CommandError(returncode=1)
        except PermissionError as pe:
            self._perm_error(pe, what, section, key)

    def read(self, args):
        section, key = self._sk(args)
        cfg = configparser.ConfigParser()
        read_config(configfile=args.configfile or ALL, config=cfg)
        value = cfg.get(section, key, fallback=None)
        if value is not None:
            log.inf(value)
        else:
            log.dbg('{} is unset'.format(args.name))
            raise CommandError(returncode=1)

    def write(self, args):
        section, key = self._sk(args)
        what = args.configfile or LOCAL
        try:
            update_config(section, key, args.value, configfile=what)
        except PermissionError as pe:
            self._perm_error(pe, what, section, key)

    def _sk(self, args):
        name_list = args.name.split(".", 1)
        if len(name_list) != 2:
            self.parser.error("name '{}' should be in the form "
                              "<section>.<key>".format(args.name))
        return name_list[0], name_list[1]

    def _perm_error(self, pe, what, section, key):
        rootp = ('; are you root/administrator?' if what in [SYSTEM, ALL]
                 else '')
        log.die("can't update {}.{}: permission denied when writing {}{}".
                format(section, key, pe.filename, rootp))



# Copyright 2018 Open Source Foundries Limited.
#
# SPDX-License-Identifier: Apache-2.0

import setuptools

with open('README.rst', 'r') as f:
    long_description = f.read()

with open('src/west/version.py', 'r') as f:
    exec(f.read())

setuptools.setup(
    name='west',
    version=__version__,
    author='Zephyr Project',
    author_email='devel@lists.zephyrproject.org',
    description='Zephyr RTOS Project meta-tool',
    long_description=long_description,
    # http://docutils.sourceforge.net/FAQ.html#what-s-the-official-mime-type-for-restructuredtext-data
    long_description_content_type="text/x-rst",
    url='https://github.com/zephyrproject-rtos/west',
    packages=setuptools.find_namespace_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],
    install_requires=[
        'colorama',
        'PyYAML',
        'pykwalify',
        'configobj',
        'setuptools>=v40.1.0',  # for find_namespace_packages
    ],
    python_requires='>=3.4',
    entry_points={'console_scripts': ('west = west.main:main',)},
)



# Copyright (c) 2018, 2019, Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0

'''West configuration file handling.

West follows Git-like conventions for configuration file locations.
There are three types of configuration file: system-wide files apply
to all users on the current machine, global files apply to the current
user, and local files apply to the current west installation.

System files:

- Linux: ``/etc/westconfig``
- macOS: ``/usr/local/etc/westconfig``
- Windows: ``%PROGRAMDATA%\\west\\config``

Global files:

- Linux: ``~/.westconfig`` or (if ``$XDG_CONFIG_HOME`` is set)
  ``$XDG_CONFIG_HOME/west/config``
- macOS: ``~/.westconfig``
- Windows: ``.westconfig`` in the user's home directory, as determined
  by os.path.expanduser.

Local files:

- Linux, macOS, Windows: ``<installation-root-directory>/.west/config``

You can override these files' locations with the ``WEST_CONFIG_SYSTEM``,
``WEST_CONFIG_GLOBAL``, and ``WEST_CONFIG_LOCAL`` environment variables.

Configuration values from later configuration files override configuration
from earlier ones. Local values have highest precedence, and system values
lowest.
'''

import configparser
import os
import pathlib
import platform
from enum import Enum

import configobj

from west.util import west_dir, WestNotFound, canon_path


# Configuration values.
#
# Initially empty, populated in read_config(). Always having this available is
# nice in case something checks configuration values before the configuration
# file has been read (e.g. the log.py functions, to check color settings, and
# tests).
config = configparser.ConfigParser(allow_no_value=True)

class ConfigFile(Enum):
    '''Types of west configuration file.

    Enumeration members:

    - SYSTEM: system level configuration shared by all users
    - GLOBAL: global or user-wide configuration
    - LOCAL: per-installation configuration
    - ALL: all three of the above, where applicable
    '''
    ALL = 1
    SYSTEM = 2
    GLOBAL = 3
    LOCAL = 4

def read_config(configfile=None, config=config, config_file=None):
    '''Read configuration files into *config*.

    :param configfile: a `west.configuration.ConfigFile`
    :param config: configuration object to read into
    :param config_file: deprecated alternative spelling for *configfile*

    Reads the files given by *configfile*, storing the values into
    the configparser.ConfigParser object *config*. If *config* is not
    given, the global `west.configuration.config` object is used.

    If *configfile* is given, only the files implied by its value are
    read. If not given, ``ConfigFile.ALL`` is used.
    '''
    if configfile is not None and config_file is not None:
        raise ValueError('use "configfile" or "config_file"; not both')
    if configfile is None:
        configfile = ConfigFile.ALL
    if config_file is not None:
        configfile = config_file
    config.read(_gather_configs(configfile), encoding='utf-8')


def update_config(section, key, value, configfile=ConfigFile.LOCAL):
    '''Sets ``section.key`` to *value* in the given configuration file.

    :param section: config section; will be created if it does not exist
    :param key: key to set in the given section
    :param value: value to set the key to
    :param configfile: `west.configuration.ConfigFile`, must not be ALL

    The destination file to write is given by *configfile*. The
    default value (``ConfigFile.LOCAL``) writes to the per-installation
    file .west/config. This function must therefore be called from a
    west installation if this default is used, or WestNotFound will be
    raised.
    '''
    if configfile == ConfigFile.ALL:
        # Not possible to update ConfigFile.ALL, needs specific conf file here.
        raise ValueError('invalid configfile: {}'.format(configfile))

    filename = _ensure_config(configfile)
    updater = configobj.ConfigObj(filename)
    if section not in updater:
        updater[section] = {}
    updater[section][key] = value
    updater.write()

def delete_config(section, key, configfile=None):
    '''Delete the option section.key from the given file or files.

    :param section: section whose key to delete
    :param key: key to delete
    :param configfile: If ConfigFile.ALL, delete section.key in all files
                       where it is set.
                       If None, delete only from the highest-precedence
                       global or local file where it is set, allowing
                       lower-precedence values to take effect again.
                       If a list of ConfigFile enumerators, delete
                       from those files.
                       Otherwise, delete from the given ConfigFile.

    Deleting the only key in a section deletes the entire section.

    If the option is not set, KeyError is raised.'''
    stop = False
    if configfile is None:
        to_check = [_location(x) for x in
                    [ConfigFile.LOCAL, ConfigFile.GLOBAL]]
        stop = True
    elif configfile == ConfigFile.ALL:
        to_check = [_location(x) for x in
                    [ConfigFile.SYSTEM, ConfigFile.GLOBAL, ConfigFile.LOCAL]]
    elif isinstance(configfile, ConfigFile):
        to_check = [_location(configfile)]
    else:
        to_check = [_location(x) for x in configfile]

    found = False
    for path in to_check:
        cobj = configobj.ConfigObj(path)
        if section not in cobj or key not in cobj[section]:
            continue

        del cobj[section][key]
        if not cobj[section].items():
            del cobj[section]
        cobj.write()
        found = True
        if stop:
            break

    if not found:
        raise KeyError('{}.{}'.format(section, key))

def _location(cfg):
    # Making this a function that gets called each time you ask for a
    # configuration file makes it respect updated environment
    # variables (such as XDG_CONFIG_HOME, PROGRAMDATA) if they're set
    # during the program lifetime. It also makes it easier to
    # monkey-patch for testing :).
    env = os.environ

    if cfg == ConfigFile.ALL:
        raise ValueError('ConfigFile.ALL has no location')
    elif cfg == ConfigFile.SYSTEM:
        if 'WEST_CONFIG_SYSTEM' in os.environ:
            return os.environ['WEST_CONFIG_SYSTEM']

        plat = platform.system()
        if plat == 'Linux':
            return '/etc/westconfig'
        elif plat == 'Darwin':
            return '/usr/local/etc/westconfig'
        elif plat == 'Windows':
            return os.path.expandvars('%PROGRAMDATA%\\west\\config')
        elif 'BSD' in plat:
            return '/etc/westconfig'
        else:
            raise ValueError('unsupported platform ' + plat)
    elif cfg == ConfigFile.GLOBAL:
        if 'WEST_CONFIG_GLOBAL' in os.environ:
            return os.environ['WEST_CONFIG_GLOBAL']
        elif platform.system() == 'Linux' and 'XDG_CONFIG_HOME' in env:
            return os.path.join(env['XDG_CONFIG_HOME'], 'west', 'config')
        else:
            return canon_path(
                os.path.join(os.path.expanduser('~'), '.westconfig'))
    elif cfg == ConfigFile.LOCAL:
        if 'WEST_CONFIG_LOCAL' in os.environ:
            return os.environ['WEST_CONFIG_LOCAL']
        else:
            # Might raise WestNotFound!
            return os.path.join(west_dir(), 'config')
    else:
        raise ValueError('invalid configuration file {}'.format(cfg))

def _gather_configs(cfg):
    # Find the paths to the given configuration files, in increasing
    # precedence order.
    ret = []

    if cfg == ConfigFile.ALL or cfg == ConfigFile.SYSTEM:
        ret.append(_location(ConfigFile.SYSTEM))
    if cfg == ConfigFile.ALL or cfg == ConfigFile.GLOBAL:
        ret.append(_location(ConfigFile.GLOBAL))
    if cfg == ConfigFile.ALL or cfg == ConfigFile.LOCAL:
        try:
            ret.append(_location(ConfigFile.LOCAL))
        except WestNotFound:
            pass

    return ret


def _ensure_config(configfile):
    # Ensure the given configfile exists, returning its path. May
    # raise permissions errors, WestNotFound, etc.
    #
    # Uses pathlib as this is hard to implement correctly without it.
    loc = _location(configfile)
    path = pathlib.Path(loc)

    if path.is_file():
        return loc

    # Create the directory. We can't use
    #     path.parent.mkdir(..., exist_ok=True)
    # in Python 3.4, so roughly emulate its behavior.
    try:
        path.parent.mkdir(parents=True)
    except FileExistsError:
        pass

    path.touch(exist_ok=True)
    return canon_path(str(path))



# Copyright (c) 2018 Open Source Foundries Limited.
#
# SPDX-License-Identifier: Apache-2.0

'''Helpers for dealing with CMake'''

from collections import OrderedDict
import os.path
import re
import subprocess
import shutil

from west import log
from west.util import quote_sh_list

__all__ = ['run_cmake', 'run_build',
           'make_c_identifier',
           'CMakeCacheEntry', 'CMakeCache']

DEFAULT_CACHE = 'CMakeCache.txt'


def run_cmake(args, quiet=False):
    '''Run cmake to (re)generate a build system'''
    cmake = shutil.which('cmake')
    if cmake is None:
        log.die('CMake is not installed or cannot be found; cannot build.')
    cmd = [cmake] + args
    kwargs = dict()
    if quiet:
        kwargs['stdout'] = subprocess.DEVNULL
        kwargs['stderr'] = subprocess.STDOUT
    log.dbg('Running CMake:', cmd, level=log.VERBOSE_VERY)
    log.dbg('As command:', quote_sh_list(cmd), level=log.VERBOSE_VERY)
    subprocess.check_call(cmd, **kwargs)


def run_build(build_directory, extra_args=(), quiet=False):
    '''Run cmake in build tool mode in `build_directory`'''
    run_cmake(['--build', build_directory] + list(extra_args), quiet=quiet)


def make_c_identifier(string):
    '''Make a C identifier from a string in the same way CMake does.
    '''
    # The behavior of CMake's string(MAKE_C_IDENTIFIER ...)  is not
    # precisely documented. This behavior matches the test case
    # that introduced the function:
    #
    # https://gitlab.kitware.com/cmake/cmake/commit/0ab50aea4c4d7099b339fb38b4459d0debbdbd85
    ret = []

    alpha_under = re.compile('[A-Za-z_]')
    alpha_num_under = re.compile('[A-Za-z0-9_]')

    if not alpha_under.match(string):
        ret.append('_')
    for c in string:
        if alpha_num_under.match(c):
            ret.append(c)
        else:
            ret.append('_')

    return ''.join(ret)


class CMakeCacheEntry:
    '''Represents a CMake cache entry.

    This class understands the type system in a CMakeCache.txt, and
    converts the following cache types to Python types:

    Cache Type    Python type
    ----------    -------------------------------------------
    FILEPATH      str
    PATH          str
    STRING        str OR list of str (if ';' is in the value)
    BOOL          bool
    INTERNAL      str OR list of str (if ';' is in the value)
    ----------    -------------------------------------------
    '''

    # Regular expression for a cache entry.
    #
    # CMake variable names can include escape characters, allowing a
    # wider set of names than is easy to match with a regular
    # expresion. To be permissive here, use a non-greedy match up to
    # the first colon (':'). This breaks if the variable name has a
    # colon inside, but it's good enough.
    CACHE_ENTRY = re.compile(
        r'''(?P<name>.*?)                               # name
         :(?P<type>FILEPATH|PATH|STRING|BOOL|INTERNAL)  # type
         =(?P<value>.*)                                 # value
        ''', re.X)

    @classmethod
    def _to_bool(cls, val):
        # Convert a CMake BOOL string into a Python bool.
        #
        #   "True if the constant is 1, ON, YES, TRUE, Y, or a
        #   non-zero number. False if the constant is 0, OFF, NO,
        #   FALSE, N, IGNORE, NOTFOUND, the empty string, or ends in
        #   the suffix -NOTFOUND. Named boolean constants are
        #   case-insensitive. If the argument is not one of these
        #   constants, it is treated as a variable."
        #
        # https://cmake.org/cmake/help/v3.0/command/if.html
        val = val.upper()
        if val in ('ON', 'YES', 'TRUE', 'Y'):
            return True
        elif val in ('OFF', 'NO', 'FALSE', 'N', 'IGNORE', 'NOTFOUND', ''):
            return False
        elif val.endswith('-NOTFOUND'):
            return False
        else:
            try:
                v = int(val)
                return v != 0
            except ValueError as exc:
                raise ValueError('invalid bool {}'.format(val)) from exc

    @classmethod
    def from_line(cls, line, line_no):
        # Comments can only occur at the beginning of a line.
        # (The value of an entry could contain a comment character).
        if line.startswith('//') or line.startswith('#'):
            return None

        # Whitespace-only lines do not contain cache entries.
        if not line.strip():
            return None

        m = cls.CACHE_ENTRY.match(line)
        if not m:
            return None

        name, type_, value = (m.group(g) for g in ('name', 'type', 'value'))
        if type_ == 'BOOL':
            try:
                value = cls._to_bool(value)
            except ValueError as exc:
                args = exc.args + ('on line {}: {}'.format(line_no, line),)
                raise ValueError(args) from exc
        elif type_ == 'STRING' or type_ == 'INTERNAL':
            # If the value is a CMake list (i.e. is a string which
            # contains a ';'), convert to a Python list.
            if ';' in value:
                value = value.split(';')

        return CMakeCacheEntry(name, value)

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __str__(self):
        fmt = 'CMakeCacheEntry(name={}, value={})'
        return fmt.format(self.name, self.value)


class CMakeCache:
    '''Parses and represents a CMake cache file.'''

    @staticmethod
    def from_build_dir(build_dir):
        return CMakeCache(os.path.join(build_dir, DEFAULT_CACHE))

    def __init__(self, cache_file):
        self.cache_file = cache_file
        self.load(cache_file)

    def load(self, cache_file):
        entries = []
        with open(cache_file, 'r') as cache:
            for line_no, line in enumerate(cache):
                entry = CMakeCacheEntry.from_line(line, line_no)
                if entry:
                    entries.append(entry)
        self._entries = OrderedDict((e.name, e) for e in entries)

    def get(self, name, default=None):
        entry = self._entries.get(name)
        if entry is not None:
            return entry.value
        else:
            return default

    def get_list(self, name, default=None):
        if default is None:
            default = []
        entry = self._entries.get(name)
        if entry is not None:
            value = entry.value
            if isinstance(value, list):
                return value
            elif isinstance(value, str):
                return [value] if value else []
            else:
                msg = 'invalid value {} type {}'
                raise RuntimeError(msg.format(value, type(value)))
        else:
            return default

    def __contains__(self, name):
        return name in self._entries

    def __getitem__(self, name):
        return self._entries[name].value

    def __setitem__(self, name, entry):
        if not isinstance(entry, CMakeCacheEntry):
            msg = 'improper type {} for value {}, expecting CMakeCacheEntry'
            raise TypeError(msg.format(type(entry), entry))
        self._entries[name] = entry

    def __delitem__(self, name):
        del self._entries[name]

    def __iter__(self):
        return iter(self._entries.values())
